################################################################################
# Module: __init__.py
# Description: Archetypal: Retrieve, construct and analyse building archetypes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

# Version of the package
__version__ = '1.2.4'
# Latest version of EnergyPlus compatible with archetypal
ep_version = '8-9-0'

# warn if a newer version of archetypal is available
from outdated import warn_if_outdated
warn_if_outdated('archetypal', __version__)

from .utils import *
from .simple_glazing import *
from .idfclass import *
from .schedule import Schedule
from .dataportal import *
from .plot import *
from .trnsys import *
from .cli import *


################################################################################
# Module: settings.py
# Description: Various settings used across the package
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import logging as lg

# locations to save data, logs, images, and cache
import archetypal

data_folder = 'data'
logs_folder = 'logs'
imgs_folder = 'images'
cache_folder = 'cache'
umitemplate = 'data/BostonTemplateLibrary.json'

# cache server responses
use_cache = False

# write log to file and/or to console
log_file = False
log_console = False
log_notebook = False
log_level = lg.INFO
log_name = 'archetypal'
log_filename = 'archetypal'

# usual idfobjects
useful_idf_objects = ['WINDOWMATERIAL:GAS', 'WINDOWMATERIAL:GLAZING',
                      'WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM', 'MATERIAL',
                      'MATERIAL:NOMASS', 'CONSTRUCTION',
                      'BUILDINGSURFACE:DETAILED',
                      'FENESTRATIONSURFACE:DETAILED', 'SCHEDULE:DAY:INTERVAL',
                      'SCHEDULE:WEEK:DAILY', 'SCHEDULE:YEAR']

# List of Available SQLite Tables
# Ref: https://bigladdersoftware.com/epx/docs/8-3/output-details-and-examples
# /eplusout.sql.html#schedules-table

available_sqlite_tables = dict(
    ComponentSizes={'PrimaryKey': ['ComponentSizesIndex'],
                    'ParseDates': []},
    ConstructionLayers={'PrimaryKey': ['ConstructionIndex'],
                        'ParseDates': []},
    Constructions={'PrimaryKey': ['ConstructionIndex'],
                   'ParseDates': []},
    Materials={'PrimaryKey': ['MaterialIndex'],
               'ParseDates': []},
    NominalBaseboardHeaters={'PrimaryKey': ['NominalBaseboardHeaterIndex'],
                             'ParseDates': []},
    NominalElectricEquipment={'PrimaryKey': ['NominalElectricEquipmentIndex'],
                              'ParseDates': []},
    NominalGasEquipment={'PrimaryKey': ['NominalGasEquipmentIndex'],
                         'ParseDates': []},
    NominalHotWaterEquipment={'PrimaryKey': ['NominalHotWaterEquipmentIndex'],
                              'ParseDates': []},
    NominalInfiltration={'PrimaryKey': ['NominalInfiltrationIndex'],
                         'ParseDates': []},
    NominalLighting={'PrimaryKey': ['NominalLightingIndex'],
                     'ParseDates': []},
    NominalOtherEquipment={'PrimaryKey': ['NominalOtherEquipmentIndex'],
                           'ParseDates': []},
    NominalPeople={'PrimaryKey': ['NominalPeopleIndex'],
                   'ParseDates': []},
    NominalSteamEquipment={'PrimaryKey': ['NominalSteamEquipmentIndex'],
                           'ParseDates': []},
    NominalVentilation={'PrimaryKey': ['NominalVentilationIndex'],
                        'ParseDates': []},
    ReportData={'PrimaryKey': ['ReportDataIndex'],
                'ParseDates': []},
    ReportDataDictionary={'PrimaryKey': ['ReportDataDictionaryIndex'],
                          'ParseDates': []},
    ReportExtendedData={'PrimaryKey': ['ReportExtendedDataIndex'],
                        'ParseDates': []},
    RoomAirModels={'PrimaryKey': ['ZoneIndex'],
                   'ParseDates': []},
    Schedules={'PrimaryKey': ['ScheduleIndex'],
               'ParseDates': []},
    Surfaces={'PrimaryKey': ['SurfaceIndex'],
              'ParseDates': []},
    SystemSizes={'PrimaryKey': ['SystemSizesIndex'],
                 'ParseDates': {'PeakHrMin': '%m/%d %H:%M:%S'}},
    Time={'PrimaryKey': ['TimeIndex'],
          'ParseDates': []},
    ZoneGroups={'PrimaryKey': ['ZoneGroupIndex'],
                'ParseDates': []},
    Zone={'PrimaryKey': ['ZoneIndex'],
          'ParseDates': []},
    ZoneLists={'PrimaryKey': ['ZoneListIndex'],
               'ParseDates': []},
    ZoneSizes={'PrimaryKey': ['ZoneSizesIndex'],
               'ParseDates': []},
    ZoneInfoZoneLists={'PrimaryKey': ['ZoneListIndex'],
                       'ParseDates': []},
    Simulations={'PrimaryKey': ['SimulationIndex'],
                 'ParseDates': {'TimeStamp': {'format': 'YMD=%Y.%m.%d %H:%M'}}},
    EnvironmentPeriods={'PrimaryKey': ['EnvironmentPeriodIndex'],
                        'ParseDates': []},
    TabularData={'PrimaryKey': ['TabularDataIndex'],
                 'ParseDates': []},
    Strings={'PrimaryKey': ['StringIndex'],
             'ParseDates': []},
    StringTypes={'PrimaryKey': ['StringTypeIndex'],
                 'ParseDates': []},
    TabularDataWithStrings={'PrimaryKey': ['TabularDataIndex'],
                            'ParseDates': []},
    Errors={'PrimaryKey': ['ErrorIndex'],
            'ParseDates': []})

# common_umi_objects
common_umi_objects = []

# default crs for when creating plots and querying databases
default_crs = {'init': 'epsg:4326'}

# unique schedule number as list
unique_schedules = []

# TRNSYS default location
trnsys_default_folder = r"C:\TRNSYS18"

# region read template - use io.BytesIO(settings.template) in code
import pkg_resources

resource_package = archetypal.__name__  # Could be any module/package name

# originBUISketchUp.idf template
resource_path = '/'.join(('templates', 'originBUISketchUp.idf'))
# Do not use os.path.join()
template_BUI = pkg_resources.resource_string(resource_package, resource_path)

# window library ('W74-lib.dat') template
resource_path = '/'.join(('templates', 'W74-lib.dat'))
# Do not use os.path.join()
template_winLib = pkg_resources.resource_string(resource_package,
                                                resource_path)

# NewFileTemplate.d18 ('W74-lib.dat') template
resource_path = '/'.join(('templates', 'NewFileTemplate.d18'))
# Do not use os.path.join()
path_template_d18 = pkg_resources.resource_filename(resource_package,
                                                resource_path)
# endregion



################################################################################
# Module: plot.py
# Description: Plot energy profiles, spatial geometries, networks, and routes
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
# OSMnx
#
# Copyright (c) 2019 Geoff Boeing https://geoffboeing.com/
#
# Part of the following code is a derivative work of the code from the OSMnx
# project, which is licensed MIT License. This code therefore is also
# licensed under the terms of the The MIT License (MIT).
################################################################################

import os
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from archetypal import settings, log
from matplotlib import cm
from matplotlib.colors import LightSource


def save_and_show(fig, ax, save, show, close, filename, file_format, dpi,
                  axis_off, extent):
    """Save a figure to disk and show it, as specified.

    Args:
        extent (str or `~matplotlib.transforms.Bbox`, optional): Bbox in
            inches. Only the given portion of the figure is saved. If
            'tight', try to figure out the tight bbox of the figure. If None,
            use savefig.bbox
        fig (matplotlib.figure.Figure): the figure
        ax (matplotlib.axes.Axes): the axes
        save (bool): whether to save the figure to disk or not
        show (bool): whether to display the figure or not
        close (bool): close the figure (only if show equals False) to prevent
            display
        filename (string): the name of the file to save
        file_format (string): the format of the file to save (e.g., 'jpg',
            'png', 'svg')
        dpi (int): the resolution of the image file if saving (Dots per inch)
        axis_off (bool): if True matplotlib axis was turned off by plot_graph so
            constrain the saved figure's extent to the interior of the axis

    Returns:
        (tuple) fig, ax
    """
    # save the figure if specified

    if save:
        start_time = time.time()

        # create the save folder if it doesn't already exist
        if not os.path.exists(settings.imgs_folder):
            os.makedirs(settings.imgs_folder)
        path_filename = os.path.join(settings.imgs_folder,
                                     os.extsep.join([filename, file_format]))

        if not isinstance(ax, (np.ndarray, list)):
            ax = [ax]
        if file_format == 'svg':
            fig.patch.set_alpha(0.)
            fig.savefig(path_filename, bbox_inches=extent, format=file_format,
                        facecolor=fig.get_facecolor(), transparent=True)
        else:
            if extent is None:
                if len(ax) == 1:
                    if axis_off:
                        for ax in ax:
                            # if axis is turned off, constrain the saved
                            # figure's extent to the interior of the axis
                            extent = ax.get_window_extent().transformed(
                                fig.dpi_scale_trans.inverted())
                else:
                    pass
            fig.savefig(path_filename, dpi=dpi, bbox_inches=extent,
                        format=file_format, facecolor=fig.get_facecolor(),
                        transparent=True)
        log('Saved the figure to disk in {:,.2f} seconds'.format(time.time() -
                                                                 start_time))

    # show the figure if specified
    if show:
        start_time = time.time()
        plt.show()
        # fig.show()
        log('Showed the plot in {:,.2f} seconds'.format(time.time() -
                                                        start_time))
    # if show=False, close the figure if close=True to prevent display
    elif close:
        plt.close()

    return fig, ax


def plot_energyprofile(energyprofile, kind='polygon', axis_off=True, cmap=None,
                       fig_height=None, fig_width=6, show=True, view_angle=-60,
                       save=False, close=False, dpi=300, file_format='png',
                       color=None, axes=None, vmin=None, vmax=None,
                       filename=None, **kwargs):
    """

    Args:
        energyprofile:
        axis_off:
        cmap:
        fig_height:
        fig_width:
        show:
        save:
        close:
        dpi:
        file_format:
        color:
        axes:
        vmin:
        vmax:
        filename:
        **kwargs:

    Returns:

    """
    if energyprofile.empty:
        warnings.warn("The EnergyProgile you are attempting to plot is "
                      "empty. Nothing has been displayed.", UserWarning)
        return axes

    import matplotlib.pyplot as plt
    # noinspection PyUnresolvedReferences
    from mpl_toolkits.mplot3d import Axes3D

    if isinstance(energyprofile.index, pd.MultiIndex):
        groups = energyprofile.groupby(level=0)
        nax = len(groups)
    else:
        nax = 1
        groups = [('unnamed', energyprofile)]

    if fig_height is None:
        fig_height = fig_width * nax

    # Set up plot
    fig, axes = plt.subplots(nax, 1, subplot_kw=dict(projection='3d'),
                             figsize=(fig_width, fig_height), dpi=dpi)
    if not isinstance(axes, np.ndarray):
        axes = [axes]

    for ax, (name, profile) in zip(axes, groups):
        values = profile.values

        vmin = values.min() if vmin is None else vmin
        vmax = values.max() if vmax is None else vmax

        facecolor = kwargs.pop('facecolor', None)
        if color is not None:
            facecolor = color

        if kind == 'polygon':
            z = values.reshape(365, 24)
            nrows, ncols = z.shape
            xs = np.linspace(0, 23, ncols)
            # y = np.linspace(0, 364, nrows)
            # The ith polygon will appear on the plane y = zs[i]
            zs = np.linspace(0, 364, nrows)
            verts = []
            for i in zs:
                ys = z[int(i), :]
                verts.append(polygon_under_graph(xs, ys))

            plot_poly_collection(ax, verts, zs,
                                 edgecolors=kwargs.get('edgecolors', None),
                                 facecolors=kwargs.get('facecolors', None),
                                 linewidths=kwargs.get('linewidths', None),
                                 cmap=cmap)
        elif kind == 'surface':
            z = values.reshape(365, 24)
            nrows, ncols = z.shape
            x = np.linspace(1, 24, ncols)
            y = np.linspace(1, 365, nrows)
            x, y = np.meshgrid(x, y)
            plot_surface(ax, x, y, z, cmap=cmap, **kwargs)
        else:
            raise NameError('plot kind "{}" is not supported'.format(kind))

        if filename is None:
            filename = 'unnamed'

        # set the extent of the figure
        ax.set_xlim3d(-1, 24)
        ax.set_xlabel('X')
        ax.set_ylim3d(-1, 365)
        ax.set_ylabel('Y')
        ax.set_zlim3d(vmin, vmax)
        ax.set_zlabel('Z')

        # configure axis appearance
        xaxis = ax.xaxis
        yaxis = ax.yaxis
        zaxis = ax.zaxis

        xaxis.get_major_formatter().set_useOffset(False)
        yaxis.get_major_formatter().set_useOffset(False)
        zaxis.get_major_formatter().set_useOffset(False)

        # if axis_off, turn off the axis display set the margins to zero and
        # point
        # the ticks in so there's no space around the plot
        if axis_off:
            ax.axis('off')
            ax.margins(0)
            ax.tick_params(which='both', direction='in')
            xaxis.set_visible(False)
            yaxis.set_visible(False)
            zaxis.set_visible(False)
            fig.canvas.draw()
        if view_angle is not None:
            ax.view_init(30, view_angle)
            ax.set_proj_type(kwargs.get('proj_type', 'persp'))
            fig.canvas.draw()
    fig, axes = save_and_show(fig=fig, ax=axes, save=save, show=show,
                              close=close, filename=filename,
                              file_format=file_format, dpi=dpi,
                              axis_off=axis_off, extent=None)
    return fig, axes


def plot_poly_collection(ax, verts, zs=None, color=None, cmap=None,
                         vmin=None, vmax=None, **kwargs):
    from matplotlib.collections import PolyCollection

    # if None in zs:
    #     zs = None

    # color=None overwrites specified facecolor/edgecolor with default color
    if color is not None:
        kwargs['color'] = color
    import matplotlib as mpl
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    poly = PolyCollection(verts, **kwargs)
    if zs is not None:
        poly.set_array(np.asarray(zs))
        poly.set_cmap(cmap)
        poly.set_clim(vmin, vmax)

    ax.add_collection3d(poly, zs=zs, zdir='y')
    # ax.autoscale_view()
    return poly


def plot_surface(ax, x, y, z, cmap=None, **kwargs):
    if cmap is None:
        cmap = cm.gist_earth

    ls = LightSource(270, 45)
    # To use a custom hillshading mode, override the built-in shading and pass
    # in the rgb colors of the shaded surface calculated from "shade".
    rgb = ls.shade(z, cmap=cm.get_cmap(cmap), vert_exag=0.1, blend_mode='soft')
    surf = ax.plot_surface(x, y, z, rstride=1, cstride=1, facecolors=rgb,
                           linewidth=0, antialiased=False, shade=False)
    return surf


def polygon_under_graph(xlist, ylist):
    """Construct the vertex list which defines the polygon filling the space
    under
    the (xlist, ylist) line graph.  Assumes the xs are in ascending order."""
    return [(xlist[0], 0.), *zip(xlist, ylist), (xlist[-1], 0.)]



################################################################################
# Module: idfclass.py
# Description: Various functions for processing of EnergyPlus models and
#              retrieving results in different forms
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
import datetime
import glob
import hashlib
import logging as lg
import multiprocessing
import os
import time
from subprocess import CalledProcessError, check_call

import eppy
import eppy.modeleditor
import pandas as pd
from eppy.EPlusInterfaceFunctions import parse_idd
from eppy.easyopen import getiddfile
from eppy.runner.run_functions import run

from archetypal import log, settings, EnergyPlusProcessError, cd


class IDF(eppy.modeleditor.IDF):
    """Wrapper over the eppy.modeleditor.IDF class

    """

    def __init__(self, *args, **kwargs):
        super(IDF, self).__init__(*args, **kwargs)
        self.schedules_dict = self.get_all_schedules()
        self.sql = None

    def run_eplus(self, weather_file=None, output_folder=None, ep_version=None,
                  output_report='sql', prep_outputs=True, **kwargs):
        """wrapper around the :func:`archetypal.run_eplus()` method.

        If weather file is defined in the IDF object, then this field is
        optional. By default, will load the sql in self.sql.

        Args:
            weather_file (str, optional): path to the EPW weather file.
            output_folder (str, optional): path to the output folder. Will
                default to the settings.cache_folder.
            ep_version (str, optional): EnergyPlus version to use for the
                simulation, eg: 8-9-0. If the idf file's version is lower than
                the chosen simulation version, then the file will be upgraded
                using :func:`archetypal.perform_transition`.
            output_report (str): 'htm' or 'sql' or None. If None, EnergyPlus
                runs but nothing is returned in the return statement.
            prep_outputs (bool or list, optional): if True, default meters and
                variables outputs are appended to the idf file. see
                :func:`prepare_outputs` for more details.
            **kwargs:

        Returns:
            The output report or the sql file loaded as a dict of DataFrames.
        """

        if not weather_file:
            weather_file = self.epw
        if not ep_version:
            ep_version = '-'.join(map(str, self.idd_version))
        eplus_file = self.idfname
        results = run_eplus(eplus_file, weather_file, output_folder, ep_version,
                            output_report, prep_outputs, **kwargs)
        if output_report != 'sql':
            # user wants something more than the sql
            return results
        else:
            # user simply wants the sql
            self.sql = results
            return results

    def add_object(self, ep_object, save=True, **kwargs):
        """Add a new object to an idf file. The function will test if the
        object exists to prevent duplicates.

        Args:
            ep_object (str): the object name to add, eg. 'OUTPUT:METER' (Must
                be in all_caps)
            **kwargs: keyword arguments to pass to other functions.

        Returns:
            eppy.modeleditor.IDF: the IDF object
        """
        # get list of objects
        objs = self.idfobjects[ep_object]  # a list
        # create new object
        new_object = self.newidfobject(ep_object, **kwargs)
        # Check if new object exists in previous list
        # If True, delete the object
        if sum([str(obj).upper() == str(new_object).upper() for obj in
                objs]) > 1:
            log('object "{}" already exists in idf file'.format(ep_object),
                lg.WARNING)
            # Remove the newly created object since the function
            # `idf.newidfobject()` automatically adds it
            self.removeidfobject(new_object)
            if not save:
                return []
        else:
            if save:
                log('object "{}" added to the idf file'.format(ep_object))
                self.save()
            else:
                # return the ep_object
                return new_object

    def get_schedule_type_limits_data_by_name(self, schedule_limit_name):
        """Returns the data for a particular 'ScheduleTypeLimits' object"""
        schedule = self.getobject('ScheduleTypeLimits'.upper(),
                                  schedule_limit_name)

        if schedule is not None:
            lower_limit = schedule['Lower_Limit_Value']
            upper_limit = schedule['Upper_Limit_Value']
            numeric_type = schedule['Numeric_Type']
            unit_type = schedule['Unit_Type']

            if schedule['Unit_Type'] == '':
                unit_type = numeric_type

            return lower_limit, upper_limit, numeric_type, unit_type
        else:
            return '', '', '', ''

    def get_schedule_data_by_name(self, sch_name, sch_type=None):
        """Returns the epbunch of a particular schedule name

        Args:
            sch_type:
        """
        if sch_type is None:
            try:
                return self.schedules_dict[sch_name]
            except:
                try:
                    schedules_dict = self.get_all_schedules()
                    return schedules_dict[sch_name]
                except KeyError:
                    raise KeyError('Unable to find schedule "{}" of type "{}" '
                                   'in idf file "{}"'.format(
                        sch_name, sch_type, self.idfname))
        else:
            return self.getobject(sch_type, sch_name)

    def get_all_schedules(self, yearly_only=False):
        """Returns all schedule ep_objects in a dict with their name as a key

        Args:
            yearly_only (bool): If True, return only yearly schedules

        Returns:
            (dict of eppy.bunch_subclass.EpBunch): the schedules with their
                name as a key
        """
        schedule_types = ['Schedule:Day:Hourly'.upper(),
                          'Schedule:Day:Interval'.upper(),
                          'Schedule:Day:List'.upper(),
                          'Schedule:Week:Daily'.upper(),
                          'Schedule:Year'.upper(),
                          'Schedule:Week:Compact'.upper(),
                          'Schedule:Compact'.upper(),
                          'Schedule:Constant'.upper(),
                          'Schedule:File'.upper()]
        if yearly_only:
            schedule_types = ['Schedule:Year'.upper(),
                              'Schedule:Compact'.upper(),
                              'Schedule:Constant'.upper(),
                              'Schedule:File'.upper()]
        scheds = {}
        for sched_type in schedule_types:
            for sched in self.idfobjects[sched_type]:
                try:
                    if sched.key.upper() in schedule_types:
                        scheds[sched.Name] = sched
                except:
                    pass
        return scheds

    def get_used_schedules(self, yearly_only=False):
        """Returns all used schedules

        Args:
            yearly_only (bool): If True, return only yearly schedules

        Returns:
            (list): the schedules names

        """
        schedule_types = ['Schedule:Day:Hourly'.upper(),
                          'Schedule:Day:Interval'.upper(),
                          'Schedule:Day:List'.upper(),
                          'Schedule:Week:Daily'.upper(),
                          'Schedule:Year'.upper(),
                          'Schedule:Week:Compact'.upper(),
                          'Schedule:Compact'.upper(),
                          'Schedule:Constant'.upper(),
                          'Schedule:File'.upper()]

        used_schedules = []
        all_schedules = self.get_all_schedules(yearly_only=yearly_only)
        for object_name in self.idfobjects:
            for object in self.idfobjects[object_name]:
                if object.key.upper() not in schedule_types:
                    for fieldvalue in object.fieldvalues:
                        try:
                            if fieldvalue in all_schedules and fieldvalue not \
                                    in used_schedules:
                                used_schedules.append(fieldvalue)
                        except:
                            pass
        return used_schedules

    @property
    def day_of_week_for_start_day(self):
        """Get day of week for start day for the first found RUNPERIOD"""
        import calendar
        day = self.idfobjects["RUNPERIOD"][0]["Day_of_Week_for_Start_Day"]

        if day.lower() == "sunday":
            return calendar.SUNDAY
        elif day.lower() == "monday":
            return calendar.MONDAY
        elif day.lower() == "tuesday":
            return calendar.TUESDAY
        elif day.lower() == "wednesday":
            return calendar.WEDNESDAY
        elif day.lower() == "thursday":
            return calendar.THURSDAY
        elif day.lower() == "friday":
            return calendar.FRIDAY
        elif day.lower() == "saturday":
            return calendar.SATURDAY
        else:
            return 0

    def building_name(self, use_idfname=False):
        if use_idfname:
            return os.path.basename(self.idfname)
        else:
            bld = self.idfobjects["BUILDING"]
            if bld is not None:
                return bld[0].Name
            else:
                return os.path.basename(self.idfname)

    def rename(self, objkey, objname, newname):
        """rename all the references to this objname

        Function comes from eppy.modeleditor and was modify to compare
        the name to rename as a lower string
        (see idfobject[idfobject.objls[findex]].lower() == objname.lower())

        Args:
            objkey (EpBunch): EpBunch we want to rename and rename all the
                occurrences where this object is in the IDF file
            objname (str): The name of the EpBunch to rename
            newname (str): New name used to rename the EpBunch

        Returns:
            theobject (EpBunch): The IDF objects renameds

        """

        refnames = eppy.modeleditor.getrefnames(self, objkey)
        for refname in refnames:
            objlists = eppy.modeleditor.getallobjlists(self, refname)
            # [('OBJKEY', refname, fieldindexlist), ...]
            for robjkey, refname, fieldindexlist in objlists:
                idfobjects = self.idfobjects[robjkey]
                for idfobject in idfobjects:
                    for findex in fieldindexlist:  # for each field
                        if idfobject[idfobject.objls[findex]].lower() == \
                                objname.lower():
                            idfobject[idfobject.objls[findex]] = newname
        theobject = self.getobject(objkey, objname)
        fieldname = [item for item in theobject.objls if item.endswith('Name')][
            0]
        theobject[fieldname] = newname
        return theobject


def object_from_idfs(idfs, ep_object, first_occurrence_only=False,
                     processors=1):
    """Takes a list of parsed IDF objects and a single ep_object and returns
    a DataFrame.

    Args:
        idfs (list of dict of IDF): List or Dict of IDF objects
        ep_object (str): EnergyPlus object eg. 'WINDOWMATERIAL:GAS' as a
            string. **Most be in all caps.**
        first_occurrence_only (bool, optional): if true, returns only the
            first occurence of the object
        processors (int, optional): specify how many processors to use for a
            parallel run

    Returns:
        pandas.DataFrame: A DataFrame

    """
    if not isinstance(idfs, (list, dict)):
        idfs = [idfs]
    container = []
    start_time = time.time()
    log('Parsing {1} objects for {0} idf files...'.format(len(idfs), ep_object))
    if isinstance(idfs, dict):
        try:
            if processors == 1:
                raise Exception('Loading objects sequentially...')
            log('Loading objects in parallel...')
            # Loading objects in parallel is actually slower at the moment,
            # so we raise an Exception
            runs = [[idf, ep_object] for idfname, idf in idfs.items()]
            import concurrent.futures
            with concurrent.futures.ProcessPoolExecutor(
                    max_workers=processors) as executor:
                container = {idfname: result for (idfname, idf), result in
                             zip(idfs.items(), executor.map(
                                 object_from_idf_pool, runs))}
        except Exception as e:
            # multiprocessing not present so pass the jobs one at a time
            log('{}'.format(e))
            container = {}
            for key, idf in idfs.items():
                # Load objects from IDF files and concatenate
                this_frame = object_from_idf(idf, ep_object)
                container[key] = this_frame

        # If keys given, construct hierarchical index using the passed keys
        # as the outermost level
        this_frame = pd.concat(container, names=['Archetype', '$id'], sort=True)
        this_frame.reset_index(inplace=True)
        this_frame.drop(columns='$id', inplace=True)
    else:
        for idf in idfs:
            # Load objects from IDF files and concatenate
            this_frame = object_from_idf(idf, ep_object)
            container.append(this_frame)
        # Concat the list of DataFrames
        this_frame = pd.concat(container)

    if first_occurrence_only:
        this_frame = this_frame.groupby('Name').first()
    this_frame.reset_index(inplace=True)
    this_frame.index.rename('$id', inplace=True)
    log('Parsed {} {} object(s) in {} idf file(s) in {:,.2f} seconds'.format(
        len(this_frame), ep_object, len(idfs),
        time.time() - start_time))
    return this_frame


def object_from_idf_pool(args):
    """Wrapper for :py:func:`object_from_idf` to use in parallel calls

    Args:
        args (list): List of arguments to pass to :func:`object_from_idf`

    Returns:
        list: A list of DataFrames

    """
    return object_from_idf(args[0], args[1])


def object_from_idf(idf, ep_object):
    """Takes one parsed IDF object and a single ep_object and returns a
    DataFrame.

    Args:
        idf (eppy.modeleditor.IDF): a parsed eppy object
        ep_object (str): EnergyPlus object eg. 'WINDOWMATERIAL:GAS' as a
            string. **Most be in all caps.**

    Returns:
        pandas.DataFrame: A DataFrame. Returns an empty DataFrame if
            ep_object is not found in file.

    """
    try:
        df = pd.concat(
            [pd.DataFrame(
                obj.fieldvalues, index=obj.fieldnames[0:len(obj.fieldvalues)]).T
             for obj in
             idf.idfobjects[ep_object]],
            ignore_index=True, sort=False)
    except ValueError:
        log(
            'ValueError: EP object "{}" does not exist in frame for idf "{}. '
            'Returning empty DataFrame"'.format(
                ep_object, idf.idfname), lg.WARNING)
        return pd.DataFrame({ep_object: []})
    else:
        return df


def load_idf(eplus_file, idd_filename=None):
    """Returns a parsed IDF object from file. If
    *archetypal.settings.use_cache* is true, then the idf object is loaded
    from cache.

    Args:
        eplus_file (str): path of the idf file.
        idd_filename (str, optional): name of the EnergyPlus IDD file. If
            None, the function tries to find it.

    Returns:
        (IDF): The parsed IDF object
    """
    # Determine version of idf file by reading the text file
    if idd_filename is None:
        idd_filename = getiddfile(get_idf_version(eplus_file))

    idf = load_idf_object_from_cache(eplus_file)

    start_time = time.time()
    if idf:
        # if found in cache, return
        log('Eppy load from cache completed in {:,.2f} seconds\n'.format(
            time.time() - start_time))
        return idf
    else:
        # Else, run eppy to load the idf objects
        idf = eppy_load(eplus_file, idd_filename)
        return idf


def eppy_load(file, idd_filename):
    """Uses package eppy to parse an idf file. Will also try to upgrade the
    idf file using the EnergyPlus Transition
    executables.

    Args:
        file (str): path of the idf file
        idd_filename: path of the EnergyPlus IDD file

    Returns:
        eppy.modeleditor.IDF: IDF object

    """
    cache_filename = hash_file(file)
    # Initiate an eppy.modeleditor.IDF object
    idf_object = None
    IDF.setiddname(idd_filename, testing=True)
    while idf_object is None:
        try:
            idf_object = IDF(file)
            # Check version of IDF file against version of IDD file
            idf_version = idf_object.idfobjects['VERSION'][
                0].Version_Identifier
            idd_version = '{}.{}'.format(idf_object.idd_version[0],
                                         idf_object.idd_version[1])

            if idf_version == idd_version:
                # if the versions fit, great!
                log('The version of the IDF file "{}",\n\t'
                    'version "{}", matched the version of EnergyPlus {},'
                    '\n\tversion "{}", used to parse it.'.format(
                    os.path.basename(file),
                    idf_version,
                    idf_object.getiddname(),
                    idd_version),
                    level=lg.DEBUG)
            else:
                # if they don't fit, upgrade file
                upgrade_idf(file)
                idd_filename = getiddfile(get_idf_version(file))
                IDF.iddname = idd_filename
        # An error could occur if the iddname is not found on the system. Try
        # to upgrade the idf file
        except Exception as e:
            log('{}'.format(e))
            log('Trying to upgrade the file instead...')
            # Try to upgrade the file
            upgrade_idf(file)
            # Get idd file for newly created and upgraded idf file
            idd_filename = getiddfile(get_idf_version(file))
            IDF.iddname = idd_filename
        else:
            # when parsing is complete, save it to disk, then return object
            save_idf_object_to_cache(idf_object, idf_object.idfname,
                                     cache_filename)
    return idf_object


def save_idf_object_to_cache(idf_object, idf_file, cache_filename=None,
                             how=None):
    """Saves the object to disk. Essentially uses the pickling functions of
    python.

    Args:
        cache_filename:
        idf_object (eppy.modeleditor.IDF): an eppy IDF object
        idf_file (str): file path of idf file
        how (str, optional): How the pickling is done. Choices are 'json' or
            'pickle'. json dump doen't quite work yet. 'pickle' will save to a
            gzip'ed file instead of a regular binary file (.dat).

    Returns:
        None

    Todo:
        * Json dump does not work yet.
    """
    # upper() can't take NoneType as input.
    if how is None:
        how = ''
    # The main function
    if settings.use_cache:
        if cache_filename is None:
            cache_filename = hash_file(idf_file)
        cache_dir = os.path.join(settings.cache_folder, cache_filename)

        # create the folder on the disk if it doesn't already exist
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        if how.upper() == 'JSON':
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'json']))
            import gzip, json
            with open(cache_fullpath_filename, 'w') as file_handle:
                json.dump({key: value.__dict__ for key, value in
                           idf_object.idfobjects.items()},
                          file_handle,
                          sort_keys=True, indent=4, check_circular=True)

        elif how.upper() == 'PICKLE':
            # create pickle and dump
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'gzip']))
            import gzip
            try:
                import cPickle as pickle
            except ImportError:
                import pickle
            start_time = time.time()
            with gzip.GzipFile(cache_fullpath_filename, 'wb') as file_handle:
                pickle.dump(idf_object, file_handle, protocol=0)
            log('Saved pickle to file in {:,.2f} seconds'.format(
                time.time() - start_time))

        else:
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'dat']))
            try:
                import cPickle as pickle
            except ImportError:
                import pickle
            start_time = time.time()
            with open(cache_fullpath_filename, 'wb') as file_handle:
                pickle.dump(idf_object, file_handle, protocol=-1)
            log('Saved pickle to file in {:,.2f} seconds'.format(
                time.time() - start_time))


def load_idf_object_from_cache(idf_file, how=None):
    """Load an idf instance from cache

    Args:
        idf_file (str): path to the idf file
        how (str, optional): How the pickling is done. Choices are 'json' or
            'pickle' or 'idf'. json dump doen't quite work yet. 'pickle' will
            load from a gzip'ed file instead of a regular binary file (.dat).
            'idf' will load from idf file saved in cache.

    Returns:
        None
    """
    # upper() can't tahe NoneType as input.
    if how is None:
        how = ''
    # The main function
    if settings.use_cache:
        cache_filename = hash_file(idf_file)
        if how.upper() == 'JSON':
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'json']))
            import json
            try:
                import cPickle as pickle
            except ImportError:
                import pickle
            start_time = time.time()
            if os.path.isfile(cache_fullpath_filename):
                with open(cache_fullpath_filename, 'rb') as file_handle:
                    idf = json.load(file_handle)
                log('Loaded "{}" from pickled file in {:,.2f} seconds'.format(
                    os.path.basename(idf_file), time.time() -
                                                start_time))
                return idf

        elif how.upper() == 'PICKLE':
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'gzip']))
            import gzip
            try:
                import cPickle as pickle
            except ImportError:
                import pickle
            start_time = time.time()
            if os.path.isfile(cache_fullpath_filename):
                with gzip.GzipFile(cache_fullpath_filename,
                                   'rb') as file_handle:
                    idf = pickle.load(file_handle)
                if idf.iddname is None:
                    idf.setiddname(getiddfile(idf.model.dt['VERSION'][0][1]))
                    # idf.read()
                log('Loaded "{}" from pickled file in {:,.2f} seconds'.format(
                    os.path.basename(idf_file), time.time() -
                                                start_time))
                return idf
        elif how.upper() == 'IDF':
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename,
                                                       'idf']))
            if os.path.isfile(cache_fullpath_filename):
                version = get_idf_version(cache_fullpath_filename, doted=True)
                iddfilename = getiddfile(version)
                idf = eppy_load(cache_fullpath_filename, iddfilename)
                return idf
        else:
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename,
                                                   os.extsep.join([
                                                       cache_filename + 'idfs',
                                                       'dat']))
            try:
                import cPickle as pickle
            except ImportError:
                import pickle
            start_time = time.time()
            if os.path.isfile(cache_fullpath_filename):
                with open(cache_fullpath_filename, 'rb') as file_handle:
                    idf = pickle.load(file_handle)
                if idf.iddname is None:
                    idf.setiddname(getiddfile(idf.model.dt['VERSION'][0][1]))
                    idf.read()
                log('Loaded "{}" from pickled file in {:,.2f} seconds'.format(
                    os.path.basename(idf_file), time.time() - start_time))
                return idf


def prepare_outputs(eplus_file, outputs=None, idd_filename=None):
    """Add additional epobjects to the idf file. Users can pass in an outputs

    Args:
        eplus_file:
        outputs (bool or list):

    Examples:
        >>> objects = [{'ep_object':'OUTPUT:DIAGNOSTICS',
        >>>             'kwargs':{'Key_1':'DisplayUnusedSchedules'}}]
        >>> prepare_outputs(eplus_file, outputs=objects)

    """

    log('first, loading the idf file')
    idf = load_idf(eplus_file, idd_filename=idd_filename)
    eplus_finename = os.path.basename(eplus_file)
    idf = {eplus_finename: idf}

    if isinstance(outputs, list):
        for output in outputs:
            idf[eplus_finename].add_object(output['ep_object'], **output[
                'kwargs'])

    # SummaryReports
    idf[eplus_finename].add_object('Output:Table:SummaryReports'.upper(),
                                   Report_1_Name='AllSummary')

    # SQL output
    idf[eplus_finename].add_object('Output:SQLite'.upper(),
                                   Option_Type='SimpleAndTabular')

    # Output variables
    idf[eplus_finename].add_object('Output:Variable'.upper(),
                                   Variable_Name='Air System Total Heating '
                                                 'Energy',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('Output:Variable'.upper(),
                                   Variable_Name='Air System Total Cooling '
                                                 'Energy',
                                   Reporting_Frequency='hourly')

    # Output meters
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='HeatRejection:EnergyTransfer',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Heating:EnergyTransfer',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Cooling:EnergyTransfer',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Heating:DistrictHeating',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Heating:Electricity',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Heating:Gas',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Cooling:DistrictCooling',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Cooling:Electricity',
                                   Reporting_Frequency='hourly')
    idf[eplus_finename].add_object('OUTPUT:METER',
                                   Key_Name='Cooling:Gas',
                                   Reporting_Frequency='hourly')


def cache_runargs(eplus_file, runargs):
    import json
    output_directory = runargs['output_directory']

    runargs.update({'run_time': datetime.datetime.now().isoformat()})
    runargs.update({'idf_file': eplus_file})
    with open(os.path.join(output_directory, 'runargs.json'), 'w') as fp:
        json.dump(runargs, fp, sort_keys=True, indent=4)


def run_eplus(eplus_file, weather_file, output_folder=None, ep_version=None,
              output_report=None, prep_outputs=False, **kwargs):
    """Run an energy plus file and return the SummaryReports Tables in a list
    of [(title, table), .....]

    Args:
        eplus_file (str): path to the idf file.
        weather_file (str): path to the EPW weather file
        output_folder (str, optional): path to the output folder. Will default
            to the settings.cache_folder.
        ep_version (str, optional): EnergyPlus version to use, eg: 8-9-0
        output_report: 'htm' or 'sql'.
        prep_outputs (bool or list, optional): if true, meters and variable
            outputs will be appended to the idf files. see
            :func:`prepare_outputs`
        **kwargs: keyword arguments to pass to other functions (see below)

    Returns:
        dict: dict of [(title, table), .....]

    Keyword Args:
        annual (bool): If True then force annual simulation (default: False)
        design_day (bool): Force design-day-only simulation (default: False)
        epmacro (bool): Run EPMacro prior to simulation (default: False)
        expandobjects (bool): Run ExpandObjects prior to simulation (default:
            False)
        readvars (bool): Run ReadVarsESO after simulation (default: False)
        output_prefix (str): Prefix for output file names
        verbose (str):
        idf : str
        output_suffix (str, optional): Suffix style for output file names
            (default: L)
                L: Legacy (e.g., eplustbl.csv)
                C: Capital (e.g., eplusTable.csv)
                D: Dash (e.g., eplus-table.csv)
        version (bool, optional): Display version information (default: False)
        verbose (str): Set verbosity of runtime messages (default: v)
            v: verbose
            q: quiet
    """
    if os.path.isfile(weather_file):
        pass
    else:
        raise FileNotFoundError('Could not find weather file: {}'.format(
            weather_file))

    # use absolute paths
    eplus_file = os.path.abspath(eplus_file)

    # <editor-fold desc="Try to get cached results">
    try:
        start_time = time.time()
        cached_run_results = get_from_cache(eplus_file, output_report,
                                            **kwargs)
    except Exception as e:
        # catch other exceptions that could occur
        raise Exception('{}'.format(e))
    else:
        if cached_run_results:
            # if cached run found, simply return it
            log('Succesfully parsed cached idf run in {:,.2f} seconds'.format(
                time.time() - start_time))
            return cached_run_results

    runs_not_found = eplus_file
    # </editor-fold>

    output_prefix = hash_file(eplus_file, kwargs)

    # <editor-fold desc="Upgrade the file version if needed">
    try:
        if ep_version:
            # replace the dots with "-"
            ep_version = ep_version.replace(".", "-")
        perform_transition(eplus_file, to_version=ep_version)
    except KeyError as e:
        log('file already upgraded to latest version "{}"'.format(e))

    # update the versionid of the file
    versionid = get_idf_version(eplus_file, doted=False)
    idd_filename = getiddfile(get_idf_version(eplus_file, doted=True))
    # </editor-fold>

    # Output folder check
    if not output_folder:
        output_folder = os.path.abspath(settings.cache_folder)
    # create the folder on the disk if it doesn't already exist
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    log('Output folder set to {}'.format(output_folder))

    # Prepare outputs e.g. sql table
    if prep_outputs:
        # Check if idf file has necessary objects (eg specific outputs)
        prepare_outputs(eplus_file, prep_outputs, idd_filename)

    if runs_not_found:
        # continue with simulation of other files
        log('no cached run for {}. Running EnergyPlus...'.format(
            os.path.basename(eplus_file)))

        start_time = time.time()
        from shutil import copyfile
        processed_runs = {}

        # used the hash of the original file (unmodified)
        filename_prefix = kwargs.get('output_prefix', output_prefix)
        epw = os.path.abspath(weather_file)
        runargs = {'idf': eplus_file,
                   'weather': epw,
                   'verbose': 'q',
                   'output_directory': os.path.join(output_folder,
                                                    filename_prefix),
                   'ep_version': versionid,
                   'output_prefix': filename_prefix,
                   'idd': idd_filename}
        runargs.update(kwargs)

        # Put a copy of the file in its cache folder and save runargs
        if not os.path.isfile(os.path.join(runargs['output_directory'],
                                           os.path.basename(eplus_file))):
            if not os.path.isdir(os.path.join(runargs['output_directory'])):
                os.mkdir(runargs['output_directory'])
            copyfile(eplus_file, os.path.join(runargs['output_directory'],
                                              os.path.basename(eplus_file)))
            cache_runargs(eplus_file, runargs.copy())

        # run the EnergyPlus Simulation
        multirunner(**runargs)

        log('Completed EnergyPlus in {:,.2f} seconds'.format(
            time.time() - start_time))

        # Return summary DataFrames
        cacheargs = {'eplus_file': eplus_file,
                     'output_folder': output_folder,
                     'output_report': output_report,
                     'filename_prefix': output_prefix,
                     **kwargs}
        cached_run_results = get_report(**cacheargs)
        return cached_run_results


def multirunner(**kwargs):
    """Wrapper for :func:`eppy.runner.run_functions.run` to be used when
    running IDF and EPW runs in parallel.

    Args:
        kwargs (dict): A dict made up of run() arguments.

    """
    try:
        run(**kwargs)
    except TypeError as e:
        log('{}'.format(e), lg.ERROR)
        raise TypeError('{}'.format(e))
    except CalledProcessError as e:
        # Get error file
        log('{}'.format(e), lg.ERROR)

        error_filename = os.path.join(kwargs['output_directory'],
                                      kwargs['output_prefix'] + 'out.err')
        if os.path.isfile(error_filename):
            with open(error_filename, 'r') as fin:
                log('\nError File for "{}" begins here...\n'.format(
                    os.path.basename(kwargs['idf'])), lg.ERROR)
                log(fin.read(), lg.ERROR)
                log('Error File for "{}" ends here...\n'.format(
                    os.path.basename(kwargs['idf'])), lg.ERROR)
            with open(error_filename, 'r') as stderr:
                raise EnergyPlusProcessError(cmd=e.cmd,
                                             idf=os.path.basename(
                                                 kwargs['idf']),
                                             stderr=stderr.read())
        else:
            log('Could not find error file', lg.ERROR)


def parallel_process(in_dict, function, processors=-1, use_kwargs=True):
    """A parallel version of the map function with a progress bar.

    Args:
        in_dict (dict-like): A dictionary to iterate over.
        function (function): A python function to apply to the elements of
            in_dict
        processors (int): The number of cores to use
        use_kwargs (bool): If True, pass the kwargs as arguments to `function`.

    Returns:
        [function(array[0]), function(array[1]), ...]

    Examples:
        >>> import archetypal as ar
        >>> files = ['tests/input_data/problematic/nat_ventilation_SAMPLE0.idf',
        >>>          'tests/input_data/regular/5ZoneNightVent1.idf']
        >>> wf = 'tests/input_data/CAN_PQ_Montreal.Intl.AP.716270_CWEC.epw'
        >>> files = ar.copy_file(files)
        >>> rundict = {file: dict(eplus_file=file, weather_file=wf,
        >>>                      ep_version=ep_version, annual=True,
        >>>                      prep_outputs=True, expandobjects=True,
        >>>                      verbose='q', output_report='sql')
        >>>           for file in files}
        >>> result = {file: ar.run_eplus(**rundict[file]) for file in files}

    """
    from tqdm import tqdm
    from concurrent.futures import ProcessPoolExecutor, as_completed

    if processors == -1:
        processors = min(len(in_dict), multiprocessing.cpu_count())

    if processors == 1:
        kwargs = {
            'desc': function.__name__,
            'total': len(in_dict),
            'unit': 'runs',
            'unit_scale': True,
            'leave': True
        }
        if use_kwargs:
            futures = {a: function(**in_dict[a]) for a in tqdm(in_dict,
                                                               **kwargs)}
        else:
            futures = {a: function(in_dict[a]) for a in tqdm(in_dict,
                                                             **kwargs)}
    else:
        with ProcessPoolExecutor(max_workers=processors) as pool:
            if use_kwargs:
                futures = {pool.submit(function, **in_dict[a]): a for a in
                           in_dict}
            else:
                futures = {pool.submit(function, in_dict[a]): a for a in
                           in_dict}

            kwargs = {
                'desc': function.__name__,
                'total': len(futures),
                'unit': 'runs',
                'unit_scale': True,
                'leave': True
            }

            # Print out the progress as tasks complete
            for f in tqdm(as_completed(futures), **kwargs):
                pass
    out = {}
    # Get the results from the futures.
    for key in futures:
        try:
            if processors > 1:
                out[futures[key]] = key.result()
            else:
                out[key] = futures[key]
        except Exception as e:
            log(str(e), lg.ERROR)
            out[futures[key]] = e
    return out


def hash_file(eplus_file, kwargs=None):
    """Simple function to hash a file and return it as a string.
    Will also hash the :py:func:`eppy.runner.run_functions.run()` arguments
    so that correct results are returned
    when different run arguments are used

    Args:
        eplus_file (str): path of the idf file
        **kwargs: keywords to pass to the hasher

    Returns:
        str: The digest value as a string of hexadecimal digits

    Todo:
        Hashing should include the external files used an idf file. For
        example, if a model
        uses a csv file as an input and that file changes, the hashing will
        currently not pickup that change. This
        could result in loading old results without the user knowing.
    """
    hasher = hashlib.md5()
    with open(eplus_file, 'rb') as afile:
        buf = afile.read()
        hasher.update(buf)
        hasher.update(
            kwargs.__str__().encode('utf-8'))  # Hashing the kwargs as well
    return hasher.hexdigest()


def get_report(eplus_file, output_folder=None, output_report='sql',
               filename_prefix=None, **kwargs):
    """Returns the specified report format (html or sql)

    Args:
        filename_prefix:
        eplus_file (str): path of the idf file
        output_folder (str, optional): path to the output folder. Will
            default to the settings.cache_folder.
        output_report: 'html' or 'sql'
        **kwargs: keyword arguments to pass to hasher.

    Returns:
        dict: a dict of DataFrames

    """
    # Hash the idf file with any kwargs used in the function
    if filename_prefix is None:
        filename_prefix = hash_file(eplus_file, kwargs)
    if output_report is None:
        return None
    elif 'htm' in output_report.lower():
        # Get the html report
        fullpath_filename = os.path.join(output_folder, filename_prefix,
                                         os.extsep.join(
                                             [filename_prefix + 'tbl', 'htm']))
        if os.path.isfile(fullpath_filename):
            return get_html_report(fullpath_filename)
        else:
            raise FileNotFoundError(
                'File "{}" does not exist'.format(fullpath_filename))

    elif 'sql' in output_report.lower():
        # Get the sql report
        fullpath_filename = os.path.join(output_folder, filename_prefix,
                                         os.extsep.join(
                                             [filename_prefix + 'out', 'sql']))
        if os.path.isfile(fullpath_filename):
            return get_sqlite_report(fullpath_filename)
        else:
            raise FileNotFoundError(
                'File "{}" does not exist'.format(fullpath_filename))
    else:
        return None


def get_from_cache(eplus_file, output_report='sql', **kwargs):
    """Retrieve a EPlus Tabulated Summary run result from the cache

    Args:
        eplus_file (str): the path of the eplus file
        output_report: 'html' or 'sql'
        **kwargs: keyword arguments to pass to other functions.

    Returns:
        dict: dict of DataFrames
    """
    if settings.use_cache:
        # determine the filename by hashing the eplus_file
        cache_filename_prefix = hash_file(eplus_file, kwargs)
        if output_report is None:
            return None
        elif 'htm' in output_report.lower():
            # Get the html report
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename_prefix,
                                                   os.extsep.join([
                                                       cache_filename_prefix
                                                       + 'tbl',
                                                       'htm']))
            if os.path.isfile(cache_fullpath_filename):
                return get_html_report(cache_fullpath_filename)

        elif 'sql' in output_report.lower():
            # get the SQL report
            cache_fullpath_filename = os.path.join(settings.cache_folder,
                                                   cache_filename_prefix,
                                                   os.extsep.join([
                                                       cache_filename_prefix
                                                       + 'out',
                                                       'sql']))
            if os.path.isfile(cache_fullpath_filename):
                # get reports from passed-in report names or from
                # settings.available_sqlite_tables if None are given
                return get_sqlite_report(cache_fullpath_filename,
                                         kwargs.get('report_tables',
                                                    settings.available_sqlite_tables))


def get_html_report(report_fullpath):
    """Parses the html Summary Report for each tables into a dictionary of
    DataFrames

    Args:
        report_fullpath (str): full path to the report file

    Returns:
        dict: dict of {title : table <DataFrame>,...}

    """
    from eppy.results import \
        readhtml  # the eppy module with functions to read the html
    with open(report_fullpath, 'r', encoding='utf-8') as cache_file:
        filehandle = cache_file.read()  # get a file handle to the html file

        cached_tbl = readhtml.titletable(
            filehandle)  # get a file handle to the html file

        log('Retrieved response from cache file "{}"'.format(
            report_fullpath))
        return summary_reports_to_dataframes(cached_tbl)


def summary_reports_to_dataframes(reports_list):
    """Converts a list of [(title, table),...] to a dict of {title: table
    <DataFrame>}. Duplicate keys must have their own unique names in the output
    dict.

    Args:
        reports_list (list): a list of [(title, table),...]

    Returns:
        dict: a dict of {title: table <DataFrame>}

    """
    results_dict = {}
    for table in reports_list:
        key = str(table[0])
        if key in results_dict:  # Check if key is already exists in
            # dictionary and give it a new name
            key = key + '_'
        df = pd.DataFrame(table[1])
        df = df.rename(columns=df.iloc[0]).drop(df.index[0])
        results_dict[key] = df
    return results_dict


def get_sqlite_report(report_file, report_tables=None):
    """Connects to the EnergyPlus SQL output file and retreives all tables

    Args:
        report_file (str): path of report file
        report_tables (list, optional): list of report table names to retreive.
        Defaults to settings.available_sqlite_tables

    Returns:
        dict: dict of DataFrames

    """
    # set list of report tables
    if not report_tables:
        report_tables = settings.available_sqlite_tables

    # if file exists, parse it with pandas' read_sql_query
    if os.path.isfile(report_file):
        import sqlite3
        # create database connection with sqlite3
        with sqlite3.connect(report_file) as conn:
            # empty dict to hold all DataFrames
            all_tables = {}
            # Iterate over all tables in the report_tables list
            for table in report_tables:
                try:
                    all_tables[table] = pd.read_sql_query(
                        "select * from {};".format(table), conn,
                        index_col=report_tables[table]['PrimaryKey'],
                        parse_dates=report_tables[table]['ParseDates'])
                except Exception as e:
                    log('no such table: {}'.format(table), lg.WARNING)

            log('SQL query parsed {} tables as DataFrames from {}'.format(
                len(all_tables), report_file))
            return all_tables


def upgrade_idf(files):
    """upgrades the idf file to the latest version. Implements the
    :func:`perform_transition` function.

    Args:
        files (str or list): path or list of paths to the idf file(s)

    Returns:

    """
    # Check if files is a str and put in a list
    if isinstance(files, str):
        files = [files]

    for file in files:
        try:
            perform_transition(file, to_version=None)
        except KeyError as e:
            log('file already upgraded to latest version "{}"'.format(e))


def perform_transition(file, to_version=None):
    """Transition program for idf version 1-0-0 to version 8-9-0.

    Args:
        file (str): path of idf file
        to_version (str): EnergyPlus version in the form "X-X-X".

    Returns:

    """
    versionid = get_idf_version(file, doted=False)[0:5]
    doted_version = get_idf_version(file, doted=True)
    iddfile = getiddfile(doted_version)
    if os.path.exists(iddfile):
        # if a E+ exists, pass
        pass
        # might be an old version of E+
    elif tuple(map(int, doted_version.split('.'))) < (8, 0):
        # else if the version is an old E+ version (< 8.0)
        iddfile = getoldiddfile(doted_version)
    vupdater_path, _ = iddfile.split('Energy+')
    # use to_version
    if to_version is None:
        # What is the latest E+ installed version
        to_version = find_eplus_installs(vupdater_path)
    if tuple(versionid.split('-')) > tuple(to_version.split('-')):
        log(
            'The version of the idf file "{}: v{}" is higher than any version '
            'of EnergyPlus installed on this machine. Please install '
            'EnergyPlus version "{}" or higher. Latest version found: '
            '{}'.format(os.path.basename(file), versionid, versionid,
                        to_version), lg.WARNING)
        return None
    ep_installation_name = os.path.abspath(os.path.dirname(getiddfile(
        to_version.replace('-', '.'))))
    vupdater_path = os.path.join(ep_installation_name, 'PreProcess',
                                 'IDFVersionUpdater')
    trans_exec = {
        '1-0-0': os.path.join(vupdater_path, 'Transition-V1-0-0-to-V1-0-1'),
        '1-0-1': os.path.join(vupdater_path, 'Transition-V1-0-1-to-V1-0-2'),
        '1-0-2': os.path.join(vupdater_path, 'Transition-V1-0-2-to-V1-0-3'),
        '1-0-3': os.path.join(vupdater_path, 'Transition-V1-0-3-to-V1-1-0'),
        '1-1-0': os.path.join(vupdater_path, 'Transition-V1-1-0-to-V1-1-1'),
        '1-1-1': os.path.join(vupdater_path, 'Transition-V1-1-1-to-V1-2-0'),
        '1-2-0': os.path.join(vupdater_path, 'Transition-V1-2-0-to-V1-2-1'),
        '1-2-1': os.path.join(vupdater_path, 'Transition-V1-2-1-to-V1-2-2'),
        '1-2-2': os.path.join(vupdater_path, 'Transition-V1-2-2-to-V1-2-3'),
        '1-2-3': os.path.join(vupdater_path, 'Transition-V1-2-3-to-V1-3-0'),
        '1-3-0': os.path.join(vupdater_path, 'Transition-V1-3-0-to-V1-4-0'),
        '1-4-0': os.path.join(vupdater_path, 'Transition-V1-4-0-to-V2-0-0'),
        '2-0-0': os.path.join(vupdater_path, 'Transition-V2-0-0-to-V2-1-0'),
        '2-1-0': os.path.join(vupdater_path, 'Transition-V2-1-0-to-V2-2-0'),
        '2-2-0': os.path.join(vupdater_path, 'Transition-V2-2-0-to-V3-0-0'),
        '3-0-0': os.path.join(vupdater_path, 'Transition-V3-0-0-to-V3-1-0'),
        '3-1-0': os.path.join(vupdater_path, 'Transition-V3-1-0-to-V4-0-0'),
        '4-0-0': os.path.join(vupdater_path, 'Transition-V4-0-0-to-V5-0-0'),
        '5-0-0': os.path.join(vupdater_path, 'Transition-V5-0-0-to-V6-0-0'),
        '6-0-0': os.path.join(vupdater_path, 'Transition-V6-0-0-to-V7-0-0'),
        '7-0-0': os.path.join(vupdater_path, 'Transition-V7-0-0-to-V7-1-0'),
        '7-1-0': os.path.join(vupdater_path, 'Transition-V7-1-0-to-V7-2-0'),
        '7-2-0': os.path.join(vupdater_path, 'Transition-V7-2-0-to-V8-0-0'),
        '8-0-0': os.path.join(vupdater_path, 'Transition-V8-0-0-to-V8-1-0'),
        '8-1-0': os.path.join(vupdater_path, 'Transition-V8-1-0-to-V8-2-0'),
        '8-2-0': os.path.join(vupdater_path, 'Transition-V8-2-0-to-V8-3-0'),
        '8-3-0': os.path.join(vupdater_path, 'Transition-V8-3-0-to-V8-4-0'),
        '8-4-0': os.path.join(vupdater_path, 'Transition-V8-4-0-to-V8-5-0'),
        '8-5-0': os.path.join(vupdater_path, 'Transition-V8-5-0-to-V8-6-0'),
        '8-6-0': os.path.join(vupdater_path, 'Transition-V8-6-0-to-V8-7-0'),
        '8-7-0': os.path.join(vupdater_path, 'Transition-V8-7-0-to-V8-8-0'),
        '8-8-0': os.path.join(vupdater_path, 'Transition-V8-8-0-to-V8-9-0'),
        '8-9-0': os.path.join(vupdater_path, 'Transition-V8-9-0-to-V9-0-0'),
        '9-0-0': os.path.join(vupdater_path, 'Transition-V9-0-0-to-V9-1-0'),
    }
    file = os.path.abspath(file)
    # store the directory we start in
    cwd = os.getcwd()
    run_dir = os.path.abspath(os.path.dirname(trans_exec[versionid]))

    # build a list of command line arguments
    if versionid == to_version:
        raise KeyError
    with cd(run_dir):
        transitions = [key for key in trans_exec
                       if tuple(map(int, key.split('-'))) < \
                       tuple(map(int, to_version.split('-')))
                       and tuple(map(int, key.split('-'))) >= \
                       tuple(map(int, versionid.split('-')))]
        for trans in transitions:
            try:
                trans_exec[trans]
            except KeyError:
                # there is no more updates to perfrom
                result = 0
            else:
                cmd = [trans_exec[trans], file]
                try:
                    check_call(cmd)
                except CalledProcessError as e:
                    # potentially catch contents of std out and put it in the
                    # error log
                    log('{}'.format(e), lg.ERROR)
                    raise

    log('Transition completed\n')
    # Clean 'idfnew' and 'idfold' files created by the transition porgram
    files_to_delete = glob.glob(os.path.dirname(file) + '/*.idfnew')
    files_to_delete.extend(glob.glob(os.path.dirname(file) + '/*.idfold'))
    files_to_delete.extend(glob.glob(os.path.dirname(
        file) + '/*.VCpErr'))  # Remove error files since logged to console
    for file in files_to_delete:
        if os.path.isfile(file):
            os.remove(file)


def find_eplus_installs(vupdater_path):
    """Finds all installed versions of EnergyPlus in the default location and
    returns the latest version number

    Args:
        vupdater_path (str): path of the current EnergyPlus install file

    Returns:
        (str): The version number of the latest E+ install

    """
    path_to_eplus, _ = vupdater_path.split('EnergyPlus')

    # Find all EnergyPlus folders
    list_eplus_dir = glob.glob(os.path.join(path_to_eplus, 'EnergyPlus*'))

    # check if any EnergyPlus install exists
    if not list_eplus_dir:
        raise Exception('No EnergyPlus installation found. Make sure '
                        'you have EnergyPlus installed. Go to '
                        'https://energyplus.net/downloads to download the '
                        'latest version of EnergyPlus.')

    # Find the most recent version of EnergyPlus installed from the version
    # number (at the end of the folder name)
    v0 = (0, 0, 0)  # Initialize the version number
    # Find the most recent version in the different folders found
    for dir in list_eplus_dir:
        version = dir[-5:]
        ver = tuple(map(int, version.split('-')))
        if ver > v0:
            v0 = ver

    return '-'.join(tuple(map(str, v0)))


def get_idf_version(file, doted=True):
    """Get idf version quickly by reading first few lines of idf file
    containing the 'VERSION' identifier

    Args:
        file (str): Absolute or relative Path to the idf file
        doted (bool, optional): Wheter or not to return the version number
        with periods or dashes eg.: 8.9 vs 8-9-0.
            Doted=False appends -0 to the end of the version number

    Returns:
        str: the version id

    """
    with open(os.path.abspath(file), 'r', encoding='latin-1') as fhandle:
        try:
            txt = fhandle.read()
            ntxt = parse_idd.nocomment(txt, '!')
            blocks = ntxt.split(';')
            blocks = [block.strip() for block in blocks]
            bblocks = [block.split(',') for block in blocks]
            bblocks1 = [[item.strip() for item in block] for block in bblocks]
            ver_blocks = [block for block in bblocks1
                          if block[0].upper() == 'VERSION']
            ver_block = ver_blocks[0]
            if doted:
                versionid = ver_block[1]
            else:
                versionid = ver_block[1].replace('.', '-') + '-0'
        except Exception as e:
            log('Version id for file "{}" cannot be found'.format(file))
            log('{}'.format(e))
            raise
        else:
            return versionid


def getoldiddfile(versionid):
    """find the IDD file of the E+ installation
    E+ version 7 and earlier have the idd in
    /EnergyPlus-7-2-0/bin/Energy+.idd """
    vlist = versionid.split('.')
    if len(vlist) == 1:
        vlist = vlist + ['0', '0']
    elif len(vlist) == 2:
        vlist = vlist + ['0']
    ver_str = '-'.join(vlist)
    eplus_exe, _ = eppy.runner.run_functions.install_paths(ver_str)
    eplusfolder = os.path.dirname(eplus_exe)
    iddfile = '{}/bin/Energy+.idd'.format(eplusfolder, )
    return iddfile



################################################################################
# Module: utils.py
# Description: Utility functions for configuration, logging
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################
# OSMnx
#
# Copyright (c) 2019 Geoff Boeing https://geoffboeing.com/
#
# Part of the following code is a derivative work of the code from the OSMnx
# project, which is licensed MIT License. This code therefore is also
# licensed under the terms of the The MIT License (MIT).
################################################################################

import datetime as dt
import json
import logging as lg
import os
import re
import sys
import unicodedata
import warnings
from collections import OrderedDict
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
from pandas.io.json import json_normalize

from archetypal import settings


def config(data_folder=settings.data_folder,
           logs_folder=settings.logs_folder,
           imgs_folder=settings.imgs_folder,
           cache_folder=settings.cache_folder,
           use_cache=settings.use_cache,
           log_file=settings.log_file,
           log_console=settings.log_console,
           log_level=settings.log_level,
           log_name=settings.log_name,
           log_filename=settings.log_filename,
           useful_idf_objects=settings.useful_idf_objects,
           umitemplate=settings.umitemplate,
           trnsys_default_folder=settings.trnsys_default_folder):
    """Configurations

    Args:
        data_folder (str): where to save and load data files
        logs_folder (str): where to write the log files
        imgs_folder (str): where to save figures
        cache_folder (str): where to save the simluation results
        use_cache (bool): if True, use a local cache to save/retrieve many of
            archetypal outputs such as EnergyPlus simulation results. This can
            save a lot of time by not calling the simulation and dataportal APIs
            repetitively for the same requests.
        log_file (bool): if true, save log output to a log file in logs_folder
        log_console (bool): if true, print log output to the console
        log_level (int): one of the logger.level constants
        log_name (str): name of the logger
        log_filename (str): name of the log file
        useful_idf_objects (list): a list of useful idf objects
        umitemplate (str): where the umitemplate is located
        trnsys_default_folder (str): root folder of TRNSYS install

    Returns:
        None
    """
    # set each global variable to the passed-in parameter value
    settings.use_cache = use_cache
    settings.cache_folder = cache_folder
    settings.data_folder = data_folder
    settings.imgs_folder = imgs_folder
    settings.logs_folder = logs_folder
    settings.log_console = log_console
    settings.log_file = log_file
    settings.log_level = log_level
    settings.log_name = log_name
    settings.log_filename = log_filename
    settings.useful_idf_objects = useful_idf_objects
    settings.umitemplate = umitemplate
    settings.trnsys_default_folder = validate_trnsys_folder(
        trnsys_default_folder)

    # if logging is turned on, log that we are configured
    if settings.log_file or settings.log_console:
        log('Configured archetypal')


def validate_trnsys_folder(trnsys_default_folder):
    """
    Args:
        trnsys_default_folder:
    """
    if sys.platform == 'win32':
        if os.path.isdir(trnsys_default_folder):
            return trnsys_default_folder
        else:
            raise ValueError('The provided TRNSYS path does not exist. Path={'
                             '}'.format(trnsys_default_folder))
    else:
        return trnsys_default_folder


def log(message, level=None, name=None, filename=None, avoid_console=False):
    """Write a message to the log file and/or print to the the console.

    Args:
        message (str): the content of the message to log
        level (int): one of the logger.level constants
        name (str): name of the logger
        filename (str): name of the log file
        avoid_console (bool): If True, don't print to console for this message
            only

    Returns:
        None
    """
    if level is None:
        level = settings.log_level
    if name is None:
        name = settings.log_name
    if filename is None:
        filename = settings.log_filename

    # if logging to file is turned on
    if settings.log_file:
        # get the current logger (or create a new one, if none), then log
        # message at requested level
        logger = get_logger(level=level, name=name, filename=filename)
        if level == lg.DEBUG:
            logger.debug(message)
        elif level == lg.INFO:
            logger.info(message)
        elif level == lg.WARNING:
            logger.warning(message)
        elif level == lg.ERROR:
            logger.error(message)

    # if logging to console is turned on, convert message to ascii and print to
    # the console
    if settings.log_console and not avoid_console:
        # capture current stdout, then switch it to the console, print the
        # message, then switch back to what had been the stdout. this prevents
        # logging to notebook - instead, it goes to console
        standard_out = sys.stdout
        sys.stdout = sys.__stdout__

        # convert message to ascii for console display so it doesn't break
        # windows terminals
        message = unicodedata.normalize('NFKD', make_str(message)).encode(
            'ascii', errors='replace').decode()
        print(message)
        sys.stdout = standard_out

        if level == lg.WARNING:
            warnings.warn(message)


def get_logger(level=None, name=None, filename=None):
    """Create a logger or return the current one if already instantiated.

    Args:
        level (int): one of the logger.level constants
        name (str): name of the logger
        filename (str): name of the log file

    Returns:
        logging.Logger: a Logger
    """

    if level is None:
        level = settings.log_level
    if name is None:
        name = settings.log_name
    if filename is None:
        filename = settings.log_filename

    logger = lg.getLogger(name)

    # if a logger with this name is not already set up
    if not getattr(logger, 'handler_set', None):

        # get today's date and construct a log filename
        todays_date = dt.datetime.today().strftime('%Y_%m_%d')
        log_filename = os.path.join(settings.logs_folder,
                                    '{}_{}.log'.format(filename, todays_date))

        # if the logs folder does not already exist, create it
        if not os.path.exists(settings.logs_folder):
            os.makedirs(settings.logs_folder)

        # create file handler and log formatter and set them up
        handler = lg.FileHandler(log_filename, encoding='utf-8')
        formatter = lg.Formatter(
            '%(asctime)s %(levelname)s %(name)s %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(level)
        logger.handler_set = True

    return logger


def make_str(value):
    """Convert a passed-in value to unicode if Python 2, or string if Python 3.

    Args:
        value (any): the value to convert to unicode/string

    Returns:
        unicode or string
    """
    try:
        # for python 2.x compatibility, use unicode
        return unicode(value)
    except NameError:
        # python 3.x has no unicode type, so if error, use str type
        return str(value)


def load_umi_template_objects(filename):
    """Reads

    Args:
        filename (str): path of template file

    Returns:
        dict: Dict of umi_objects
    """
    with open(filename) as f:
        umi_objects = json.load(f)
    return umi_objects


def umi_template_object_to_dataframe(umi_dict, umi_object):
    """Returns flattened DataFrame of umi_objects

    Args:
        umi_dict (dict): dict of umi objects
        umi_object (str): umi_object name

    Returns:
        pandas.DataFrame: flattened DataFrame of umi_objects
    """
    return json_normalize(umi_dict[umi_object])


def get_list_of_common_umi_objects(filename):
    """Returns list of common umi objects

    Args:
        filename (str): path to umi template file

    Returns:
        dict: Dict of common umi objects
    """
    umi_objects = load_umi_template(filename)
    components = OrderedDict()
    for umi_dict in umi_objects:
        for x in umi_dict:
            components[x] = umi_dict[x].columns.tolist()
    return components


def newrange(previous, following):
    """Takes the previous DataFrame and calculates a new Index range. Returns a
    DataFrame with a new index

    Args:
        previous (pandas.DataFrame): previous DataFrame
        following (pandas.DataFrame): follwoing DataFrame

    Returns:
        pandas.DataFrame: DataFrame with an incremented new index
    """
    if not previous.empty:
        from_index = previous.iloc[[-1]].index.values + 1
        to_index = from_index + len(following)

        following.index = np.arange(from_index, to_index)
        following.rename_axis('$id', inplace=True)
        return following
    else:
        # If previous dataframe is empty, return the orginal DataFrame
        return following


def type_surface(row):
    """Takes a boundary and returns its corresponding umi-type

    Args:
        row:

    Returns:
        str: The umi-type of boundary
    """

    # Floors
    if row['Surface_Type'] == 'Floor':
        if row['Outside_Boundary_Condition'] == 'Surface':
            return 3
        if row['Outside_Boundary_Condition'] == 'Ground':
            return 2
        if row['Outside_Boundary_Condition'] == 'Outdoors':
            return 4
        else:
            return np.NaN

    # Roofs & Ceilings
    if row['Surface_Type'] == 'Roof':
        return 1
    if row['Surface_Type'] == 'Ceiling':
        return 3
    # Walls
    if row['Surface_Type'] == 'Wall':
        if row['Outside_Boundary_Condition'] == 'Surface':
            return 5
        if row['Outside_Boundary_Condition'] == 'Outdoors':
            return 0
    return np.NaN


def label_surface(row):
    """Takes a boundary and returns its corresponding umi-Category

    Args:
        row:
    """
    # Floors
    if row['Surface_Type'] == 'Floor':
        if row['Outside_Boundary_Condition'] == 'Surface':
            return 'Interior Floor'
        if row['Outside_Boundary_Condition'] == 'Ground':
            return 'Ground Floor'
        if row['Outside_Boundary_Condition'] == 'Outdoors':
            return 'Exterior Floor'
        else:
            return 'Other'

    # Roofs & Ceilings
    if row['Surface_Type'] == 'Roof':
        return 'Roof'
    if row['Surface_Type'] == 'Ceiling':
        return 'Interior Floor'
    # Walls
    if row['Surface_Type'] == 'Wall':
        if row['Outside_Boundary_Condition'] == 'Surface':
            return 'Partition'
        if row['Outside_Boundary_Condition'] == 'Outdoors':
            return 'Facade'
    return 'Other'


def layer_composition(row):
    """Takes in a series with $id and thickness values and return an array of
    dict of the form {'Material': {'$ref': ref}, 'thickness': thickness} If
    thickness is 'nan', it returns None.

    Returns (list): List of dicts

    Args:
        row (pandas.Series): a row
    """
    array = []
    ref = row['$id', 'Outside_Layer']
    thickness = row['Thickness', 'Outside_Layer']
    if np.isnan(ref):
        pass
    else:
        array.append({'Material': {'$ref': str(int(ref))},
                      'Thickness': thickness})
        for i in range(2, len(row['$id']) + 1):
            ref = row['$id', 'Layer_{}'.format(i)]
            if np.isnan(ref):
                pass
            else:
                thickness = row['Thickness', 'Layer_{}'.format(i)]
                array.append(
                    {'Material': {'$ref': str(int(ref))},
                     'Thickness': thickness})
        return array


def schedule_composition(row):
    """Takes in a series with $id and \*_ScheduleDay_Name values and return an
    array of dict of the form {'$ref': ref}

    Args:
        row (pandas.Series): a row

    Returns:
        list: list of dicts
    """
    # Assumes 7 days
    day_schedules = []
    days = ['Monday_ScheduleDay_Name',
            'Tuesday_ScheduleDay_Name',
            'Wednesday_ScheduleDay_Name',
            'Thursday_ScheduleDay_Name',
            'Friday_ScheduleDay_Name',
            'Saturday_ScheduleDay_Name',
            'Sunday_ScheduleDay_Name']  # With weekends last (as defined in
    # umi-template)
    # Let's start with the `Outside_Layer`
    for day in days:
        try:
            ref = row['$id', day]
        except:
            pass
        else:
            day_schedules.append({'$ref': str(int(ref))})
    return day_schedules


def year_composition(row):
    """Takes in a series with $id and ScheduleWeek_Name_{} values and return an
    array of dict of the form {'FromDay': fromday, 'FromMonth': frommonth,
    'Schedule': {'$ref': int( ref)}, 'ToDay': today, 'ToMonth': tomonth}

    Args:
        row (pandas.Series): a row

    Returns:
        list: list of dicts
    """
    parts = []
    for i in range(1, 26 + 1):
        try:
            ref = row['$id', 'ScheduleWeek_Name_{}'.format(i)]
        except:
            pass
        else:
            if ~np.isnan(ref):
                fromday = row['Schedules', 'Start_Day_{}'.format(i)]
                frommonth = row['Schedules', 'Start_Month_{}'.format(i)]
                today = row['Schedules', 'End_Day_{}'.format(i)]
                tomonth = row['Schedules', 'End_Month_{}'.format(i)]

                parts.append({'FromDay': fromday,
                              'FromMonth': frommonth,
                              'Schedule': {'$ref': str(int(ref))},
                              'ToDay': today,
                              'ToMonth': tomonth})
    return parts


def date_transform(date_str):
    """Simple function transforming one-based hours (1->24) into zero-based
    hours (0->23)

    Args:
        date_str (str): a date string of the form 'HH:MM'

    Returns:
        datetime.datetime: datetime object
    """
    if date_str[0:2] != '24':
        return datetime.strptime(date_str, '%H:%M') - timedelta(hours=1)
    return datetime.strptime('23:00', '%H:%M')


def weighted_mean(series, df, weighting_variable):
    """Compute the weighted average while ignoring NaNs. Implements
    :func:`numpy.average`.

    Args:
        series (pandas.Series):
        df (pandas.DataFrame):
        weighting_variable (str or list or tuple): Weight name to use in

    Returns:
        numpy.ndarray: the weighted average
    """
    # get non-nan values
    index = ~np.isnan(series.values.astype('float'))

    # Returns weights. If multiple `weighting_variable`, df.prod will take care
    # of multipling them together.
    if not isinstance(weighting_variable, list):
        weighting_variable = [weighting_variable]
    try:
        weights = df.loc[series.index, weighting_variable].astype('float').prod(
            axis=1)
    except Exception:
        raise

    # Try to average
    try:
        wa = np.average(series[index].astype('float'), weights=weights[index])
    except ZeroDivisionError:
        log('Cannot aggregate empty series {}'.format(series.name), lg.WARNING)
        return np.NaN
    except Exception:
        raise
    else:
        return wa


def top(series, df, weighting_variable):
    """Compute the highest ranked value weighted by some other variable. Implements
        :func:`pandas.DataFrame.nlargest`.

    Args:
        series (pandas.Series): the *series* on which to compute the ranking.
        df (pandas.DataFrame): the *df* containing weighting variables.
        weighting_variable (str or list or tuple): Name of weights to use in
            *df*. If multiple values given, the values are multiplied together.

    Returns:
        numpy.ndarray: the weighted top ranked variable
    """
    # Returns weights. If multiple `weighting_variable`, df.prod will take care
    # of multipling them together.
    if not isinstance(series, pd.Series):
        raise TypeError('"top()" only works on Series, '
                        'not DataFrames\n{}'.format(series))

    if not isinstance(weighting_variable, list):
        weighting_variable = [weighting_variable]

    try:
        idx_ = df.loc[series.index].groupby(series.name).apply(
            lambda x: safe_prod(x, df, weighting_variable)
        )
        if not idx_.empty:
            idx = idx_.nlargest(1).index
        else:
            log('No such names "{}"'.format(series.name))
            return np.NaN
    except KeyError:
        log('Cannot aggregate empty series {}'.format(series.name), lg.WARNING)
        return np.NaN
    except Exception:
        raise
    else:
        if idx.isnull().any():
            return np.NaN
        else:
            return pd.to_numeric(idx, errors='ignore').values[0]


def safe_prod(x, df, weighting_variable):
    """
    Args:
        x:
        df:
        weighting_variable:
    """
    df_ = df.loc[x.index, weighting_variable]
    if not df_.empty:
        return df_.astype('float').prod(axis=1).sum()
    else:
        return 0


def copy_file(files, where=None):
    """Handles a copy of test idf files

    Args:
        files:
        where:
    """
    import shutil, os
    if isinstance(files, str):
        files = [files]
    files = {os.path.basename(k): k for k in files}

    # defaults to cache folder
    if where is None:
        where = settings.cache_folder

    for file in files:
        dst = os.path.join(where, file)
        output_folder = where
        if not os.path.isdir(output_folder):
            os.makedirs(output_folder)
        shutil.copyfile(files[file], dst)
        files[file] = dst

    return list(files.values())


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class EnergyPlusProcessError(Error):
    """EnergyPlus Process call error"""

    def __init__(self, cmd, stderr, idf=None):
        """
        Args:
            cmd:
            stderr:
            idf:
        """
        self.cmd = cmd
        self.idf = idf
        self.stderr = stderr

    def __str__(self):
        """Override that only returns the stderr"""
        msg = ':\n'.join([self.idf, self.stderr])
        return msg


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, new_path):
        """
        Args:
            new_path:
        """
        self.newPath = os.path.expanduser(new_path)

    def __enter__(self):
        self.savedPath = os.getcwd()
        if os.path.isdir(self.newPath):
            os.chdir(self.newPath)
        else:
            os.mkdir(self.newPath)
            os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        """
        Args:
            etype:
            value:
            traceback:
        """
        os.chdir(self.savedPath)


def rmse(data, targets):
    """calculate rmse with target values

    Args:
        data:
        targets:
    """
    y = piecewise(data)
    predictions = y
    error = np.sqrt(np.mean((predictions - targets) ** 2))
    return error


def piecewise(data):
    """returns a piecewise function from an array of the form [hour1, hour2,
    ..., value1, value2, ...]

    Args:
        data:
    """
    nb = int(len(data) / 2)
    bins = data[0: nb]
    sf = data[nb:]
    x = np.linspace(0, 8760, 8760)
    # build condition array
    conds = [x < bins[0]]
    conds.extend([np.logical_and(x >= i, x < j) for i, j in zip(bins[0:],
                                                                bins[1:])])
    # build function array. This is the value of y when the condition is met.
    funcs = sf
    y = np.piecewise(x, conds, funcs)
    return y


def checkStr(datafile, string):
    """Find the last occurrence of a string and return its line number

    Returns: the list index containing the string

    Args:
        datafile (list-like): a list-like object
        string (str): the string to find in the txt file
    """
    value = []
    count = 0
    for line in datafile:
        count = count + 1
        match = re.search(string, str(line))
        if match:
            return count
            break


def write_lines(file_path, lines):
    """Delete file if exists, then write lines in it

    Args:
        file_path (str): path of the file
        lines (list of str): lines to be written in file
    """
    # Delete temp file if exists
    if os.path.exists(file_path):
        os.remove(file_path)
    # Save lines in temp file
    temp_idf_file = open(file_path, "w+")
    for line in lines:
        temp_idf_file.write("%s" % line)
    temp_idf_file.close()


def load_umi_template(json_template):
    """
    Args:
        json_template: Absolute or relative filepath to an umi json_template

    Returns:
        pandas.DataFrame: 17 DataFrames, one for each component groups
    """
    if os.path.isfile(json_template):
        with open(json_template) as f:
            dicts = json.load(f, object_pairs_hook=OrderedDict)

            return [{key: json_normalize(value)} for key, value in
                    dicts.items()]
    else:
        raise ValueError('File {} does not exist'.format(json_template))


def check_unique_name(first_letters, count, name, unique_list, suffix=False):
    """Making sure new_name does not already exist

    Args:
        first_letters (str): string at the beginning of the name, giving a hint
            on what the variable is.
        count (int): increment to create a unique id in the name.
        name (str): name that was just created. To be verified that it is unique
            in this function.
        unique_list (list): list where unique names are stored.
        suffix (bool):

    Returns:
        new_name (str): name that is unique
    """
    if suffix:
        while name in unique_list:
            count += 1
            end_count = '%03d' % count
            name = name[:-3] + end_count
    else:
        while name in unique_list:
            count += 1
            end_count = '%06d' % count
            name = first_letters + '_' + end_count

    return name, count


def angle(v1, v2, acute=True):
    """Calculate the angle between 2 vectors

    Args:
        v1 (Vector3D): vector 1
        v2 (Vector3D): vector 2
        acute (bool): If True, give the acute angle, else gives the obtuse one.

    Returns:
        angle (float): angle between the 2 vectors in degree
    """
    angle = np.arccos(
        np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))
    if (acute == True):
        return angle
    else:
        return 2 * np.pi - angle



################################################################################
# Module: dataportal.py
# Description: Various functions to acquire building archetype data using
#              available APIs
# License: MIT, see full license in LICENSE.txt
# Web: https://github.com/samuelduchesne/archetypal
################################################################################

import hashlib
import io
import json
import logging as lg
import os
import re
import time
import zipfile

import pandas as pd
import pycountry as pycountry
import requests
from archetypal import log, settings, make_str

# scipy and sklearn are optional dependencies for faster nearest node search
try:
    from osgeo import gdal
except ImportError as e:
    gdal = None


def tabula_available_buildings(code_country='France'):
    """Returns all available building types for a specific country.

    Args:
        code_country:
    """
    # Check code country
    if code_country.upper() not in ['AT', 'BA', 'BE', 'BG', 'CY', 'CZ', 'DE',
                                    'DK', 'ES', 'FR', 'GB', 'GR', 'HU', 'IE',
                                    'IT', 'NL', 'NO', 'PL', 'RS', 'SE', 'SI']:
        code_country = pycountry.countries.get(name=code_country)
        if code_country is not None:
            code_country = code_country.alpha_2
        else:
            raise ValueError('Country name {} is invalid'.format(code_country))
    data = {'code_country': code_country}
    json_response = tabula_api_request(data, table='all-country')

    # load data
    df = pd.DataFrame(json_response)
    df = df.data.apply(pd.Series)
    return df


def tabula_api_request(data, table='detail'):
    """Send a request to the TABULA API via HTTP GET and return the JSON
    response.

    Args:
        data (dict): dictionnary of query attributes. with table='all-country',
            data expects 'code_country'. with table='detail', data expects
            'buildingtype', 'suffix', and 'variant'.
        table (str): the server-table to query. 'detail' or 'all-country'
    """
    # Prepare URL
    if table == 'all-country':
        codehex = str(
            int(hashlib.md5(data['code_country'].encode('utf-8')).hexdigest(),
                16))[0:13]
        url_base = ('http://webtool.building-typology.eu/data/matrix/building'
                    '/{0}/p/0/o/0/l/10/dc/{1}')
        prepared_url = url_base.format(data['code_country'], codehex)

    elif table == 'detail':
        buildingtype = '.'.join(s for s in data['buildingtype'])
        suffix = '.'.join(s for s in data['suffix'])
        bldname = buildingtype + '.' + suffix
        hexint = hashlib.md5(bldname.encode('utf-8')).hexdigest()[0:13]
        url_base = ('http://webtool.building-typology.eu/data/adv/building'
                    '/detail/{0}/bv/{1}/dc/{2}')
        prepared_url = url_base.format(bldname, data['variant'], hexint)

    else:
        raise ValueError('server-table name "{}" invalid'.format(table))

    # First, try to get the cached resonse from file
    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json
    else:
        # if this URL is not already in the cache, request it
        response = requests.get(prepared_url)
        if response.status_code == 200:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            elif not response_json['success']:
                raise ValueError('The query "{}" returned no results'.format(
                    prepared_url), lg.WARNING)
            save_to_cache(prepared_url, response_json)
            return response_json
        else:
            # Handle some server errors
            pass


def tabula_building_details_sheet(code_building=None, code_country='FR',
                                  code_typologyregion='N',
                                  code_buildingsizeclass='SFH',
                                  code_construcionyearclass=1,
                                  code_additional_parameter='Gen',
                                  code_type='ReEx',
                                  code_num=1, code_variantnumber=1):
    """
    How to format ``code_building``:
        Format the :attr:`code_building` string as such::

            "AT.MT.AB.02.Gen.ReEx.001.001"
             |  |  |  |   |   |    |   |__code_variantnumber
             |  |  |  |   |   |    |______code_num
             |  |  |  |   |   |___________code_type
             |  |  |  |   |_______________code_additional_parameter
             |  |  |  |___________________code_construcionyearclass
             |  |  |______________________code_buildingsizeclass
             |  |_________________________code_typologyregion
             |____________________________code_country

    Args:
        code_building (str): Whole building code. See How to format ``code_building``
        code_country (str): Country name or International Country Code (ISO
            3166-1-alpha-2 code). Input as 'France' will work equally as 'FR'.
        code_typologyregion (str): N for national; otherwise specific codes
            representing regions in a given country
        code_buildingsizeclass (str): 4 standardized classes: 'SFH':
        code_construcionyearclass (int or str): allocation of time bands to
            classes. Defined nationally (according to significant changes in
            construction technologies, building codes or available statistical
            data
        code_additional_parameter (str): 1 unique category. Defines the generic
            (or basic) typology matrix so that each residential building of a
            given country can be assigned to one generic type. A further
            segmentation in subtypes is possible and can be indicated by a
            specific code. Whereas the generic types must comprise the whole
            building stock the total of subtypes must be comprehensive. e.g.
            'HR' (highrises), 'TFrame' (timber frame), 'Semi' (semi-detached)
        code_type: “ReEx” is a code for “real example” and “SyAv” for
            “Synthetical Average”
        code_num: TODO: What is this paramter?
        code_variantnumber: the energy performance level 1, 2 and 3. 1: minimum
            requirements, 2: improved and 3: ambitious or NZEB standard (assumed
            or announced level of Nearly Zero-Energy Buildings)

    Returns:
        pandas.DataFrame: The DataFrame from the
    """
    # Parse builsing_code
    if code_building is not None:
        try:
            code_country, code_typologyregion, code_buildingsizeclass, \
            code_construcionyearclass, \
            code_additional_parameter, code_type, code_num, \
            code_variantnumber = code_building.split('.')
        except ValueError:
            msg = (
                'the query "{}" is missing a parameter. Make sure the '
                '"code_building" has the form: '
                'AT.MT.AB.02.Gen.ReEx.001.001').format(code_building)
            log(msg, lg.ERROR)
            raise ValueError(msg)

    # Check code country
    if code_country.upper() not in ['AT', 'BA', 'BE', 'BG', 'CY', 'CZ', 'DE',
                                    'DK', 'ES', 'FR', 'GB', 'GR', 'HU', 'IE',
                                    'IT', 'NL', 'NO', 'PL', 'RS', 'SE', 'SI']:
        code_country = pycountry.countries.get(name=code_country)
        if code_country is not None:
            # if country is valid, return ISO 3166-1-alpha-2 code
            code_country = code_country.alpha_2
        else:
            raise ValueError('Country name {} is invalid'.format(code_country))

    # Check code_buildingsizeclass
    if code_buildingsizeclass.upper() not in ['SFH', 'TH', 'MFH', 'AB']:
        raise ValueError(
            'specified code_buildingsizeclass "{}" not supported. Available '
            'values are "SFH", "TH", '
            '"MFH" or "AB"')
    # Check numericals
    if not isinstance(code_construcionyearclass, str):
        code_construcionyearclass = str(code_construcionyearclass).zfill(2)

    if not isinstance(code_num, str):
        code_num = str(code_num).zfill(3)

    if not isinstance(code_variantnumber, str):
        code_variantnumber = str(code_variantnumber).zfill(3)

    # prepare data
    data = {'buildingtype': [code_country, code_typologyregion,
                             code_buildingsizeclass, code_construcionyearclass,
                             code_additional_parameter],
            'suffix': [code_type, code_num],
            'variant': code_variantnumber}
    json_response = tabula_api_request(data, table='detail')

    if json_response is not None:
        log('')
        # load data
        df = pd.DataFrame(json_response)
        df = df.data.apply(pd.Series)

        # remove html tags from labels
        df.label = df.label.str.replace('<[^<]+?>', ' ')
        return df
    else:
        raise ValueError('No data found in TABULA matrix with query:"{}"\nRun '
                         'archetypal.dataportal.tabula_available_buildings() '
                         'with country code "{}" to get list of possible '
                         'building types'
                         ''.format('.'.join(s for s in data['buildingtype']),
                                   code_country))


def tabula_system(code_country, code_boundarycond='SUH', code_variantnumber=1):
    """
    Args:
        code_country:
        code_boundarycond:
        code_variantnumber:
    """
    # Check code country
    if code_country.upper() not in ['AT', 'BA', 'BE', 'BG', 'CY', 'CZ', 'DE',
                                    'DK', 'ES', 'FR', 'GB', 'GR', 'HU', 'IE',
                                    'IT', 'NL', 'NO', 'PL', 'RS', 'SE', 'SI']:
        code_country = pycountry.countries.get(name=code_country)
        if code_country is not None:
            # if country is valid, return ISO 3166-1-alpha-2 code
            code_country = code_country.alpha_2
        else:
            raise ValueError('Country name {} is invalid')

    # Check code_buildingsizeclass
    if code_boundarycond.upper() not in ['SUH', 'MUH']:
        raise ValueError(
            'specified code_boundarycond "{}" not valid. Available values are '
            '"SUH" (Single Unit Houses) '
            'and "MUH" (Multi-unit Houses)')

    # Check code variant number
    if not isinstance(code_variantnumber, str):
        code_variantnumber = str(code_variantnumber).zfill(2)

    # prepare data
    data = {'systype': [code_country, code_boundarycond, code_variantnumber]}
    json_response = tabula_system_request(data)

    if json_response is not None:
        log('')
        # load data
        df = pd.DataFrame(json_response)
        return df.data.to_frame()
    else:
        raise ValueError('No data found in TABULA matrix with query:"{}"\nRun '
                         'archetypal.dataportal.tabula_available_buildings() '
                         'with country code "{}" to get list of possible '
                         'building types'
                         ''.format('.'.join(s for s in data['systype']),
                                   code_country))


def tabula_system_request(data):
    """Returns:

    Examples:
        'http://webtool.building-typology.eu/data/matrix/system/detail/IT.SUH.01/dc/1546889637169'

    Args:
        data (dict): prepared data for html query
    """
    system = '.'.join(s for s in data['systype'])
    hexint = hashlib.md5(system.encode('utf-8')).hexdigest()[0:13]

    log('quering system type {}'.format(system))
    prepared_url = 'http://webtool.building-typology.eu/data/matrix/system' \
                   '/detail/{0}/dc/{1}'.format(
        system, hexint)

    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json

    else:
        # if this URL is not already in the cache, pause, then request it
        response = requests.get(prepared_url)

        try:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            save_to_cache(prepared_url, response_json)
        except Exception:
            # Handle some server errors
            pass
        else:
            return response_json


def get_from_cache(url):
    """
    Args:
        url:
    """
    # if the tool is configured to use the cache
    if settings.use_cache:
        # determine the filename by hashing the url
        filename = hashlib.md5(url.encode('utf-8')).hexdigest()

        cache_path_filename = os.path.join(settings.cache_folder,
                                           os.extsep.join([filename, 'json']))
        # open the cache file for this url hash if it already exists, otherwise
        # return None
        if os.path.isfile(cache_path_filename):
            with io.open(cache_path_filename, encoding='utf-8') as cache_file:
                response_json = json.load(cache_file)
            log('Retrieved response from cache file "{}" for URL "{}"'.format(
                cache_path_filename, url))
            return response_json


def save_to_cache(url, response_json):
    """
    Args:
        url:
        response_json:
    """
    if settings.use_cache:
        if response_json is None:
            log('Saved nothing to cache because response_json is None')
        else:
            # create the folder on the disk if it doesn't already exist
            if not os.path.exists(settings.cache_folder):
                os.makedirs(settings.cache_folder)

            # hash the url (to make filename shorter than the often extremely
            # long url)
            filename = hashlib.md5(url.encode('utf-8')).hexdigest()
            cache_path_filename = os.path.join(settings.cache_folder,
                                               os.extsep.join(
                                                   [filename, 'json']))
            # dump to json, and save to file
            json_str = make_str(json.dumps(response_json))
            with io.open(cache_path_filename, 'w',
                         encoding='utf-8') as cache_file:
                cache_file.write(json_str)

            log('Saved response to cache file "{}"'.format(cache_path_filename))


def openei_api_request(data, pause_duration=None, timeout=180,
                       error_pause_duration=None):
    """
    Args:
        data (dict or OrderedDict): key-value pairs of parameters to post to the
            API
        pause_duration:
        timeout (int): how long to pause in seconds before requests, if None,
            will query API status endpoint to find when next slot is available
        error_pause_duration (int): the timeout interval for the requests
            library

    Returns:
        dict
    """
    # define the Overpass API URL, then construct a GET-style URL as a string to
    # hash to look up/save to cache
    url = ' https://openei.org/services/api/content_assist/recommend'
    prepared_url = requests.Request('GET', url, params=data).prepare().url
    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json


def nrel_api_cbr_request(data):
    """
    Notes:
        For a detailed description of data arguments, visit
        https://developer.nrel.gov/docs/buildings/commercial-building
        -resource-database-v1/resources/

    Examples:
        >>> import archetypal as ar
        >>> ar.dataportal.nrel_api_cbr_request({'s': 'Commercial'
        >>> 'Reference', 'api_key': 'oGZdX1nhars1cTJYTm7M9T12T1ZOvikX9pH0Zudq'})

    Args:
        data: a dict of

    Returns:
        dict: the json response
    """
    # define the Overpass API URL, then construct a GET-style URL as a string to
    # hash to look up/save to cache
    url = 'https://developer.nrel.gov/api/commercial-building-resources/v1' \
          '/resources.json'
    prepared_url = requests.Request('GET', url, params=data).prepare().url
    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json

    else:
        start_time = time.time()
        log('Getting from {}, "{}"'.format(url, data))
        response = requests.get(prepared_url)
        # if this URL is not already in the cache, pause, then request it
        # get the response size and the domain, log result
        size_kb = len(response.content) / 1000.
        domain = re.findall(r'//(?s)(.*?)/', url)[0]
        log('Downloaded {:,.1f}KB from {}'
            ' in {:,.2f} seconds'.format(size_kb, domain,
                                         time.time() - start_time))

        try:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            save_to_cache(prepared_url, response_json)
        except Exception:
            # deal with response satus_code here
            log(
                'Server at {} returned status code {} and no JSON data.'.format(
                    domain,
                    response.status_code),
                level=lg.ERROR)
        else:
            return response_json


def nrel_bcl_api_request(data):
    """Send a request to the Building Component Library API via HTTP GET and
    return the JSON response.

    Args:
        data (dict or OrderedDict): key-value pairs of parameters to post to the
            API

    Returns:
        dict
    """
    try:
        kformat = data.pop('format')  # json or xml
        keyword = data.pop('keyword')
    except KeyError:
        url = 'https://bcl.nrel.gov/api/search/'
    else:
        url = 'https://bcl.nrel.gov/api/search/{}.{}'.format(keyword, kformat)
    prepared_url = requests.Request('GET', url, params=data).prepare().url
    print(prepared_url)
    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json

    else:
        start_time = time.time()
        log('Getting from {}, "{}"'.format(url, data))
        response = requests.get(prepared_url)
        # if this URL is not already in the cache, pause, then request it
        # get the response size and the domain, log result
        size_kb = len(response.content) / 1000.
        domain = re.findall(r'//(?s)(.*?)/', url)[0]
        log('Downloaded {:,.1f}KB from {}'
            ' in {:,.2f} seconds'.format(size_kb, domain,
                                         time.time() - start_time))

        try:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            save_to_cache(prepared_url, response_json)
        except Exception:
            # deal with response satus_code here
            log(
                'Server at {} returned status code {} and no JSON data.'.format(
                    domain,
                    response.status_code),
                level=lg.ERROR)
            return response.content
        else:
            return response_json


def stat_can_request(data):
    """
    Args:
        data:
    """
    prepared_url = 'https://www12.statcan.gc.ca/rest/census-recensement' \
                   '/CPR2016.{type}?lang={lang}&dguid={dguid}&topic=' \
                   '{topic}&notes={notes}'.format(
        type=data.get('type', 'json'),
        lang=data.get('land', 'E'),
        dguid=data.get('dguid', '2016A000011124'),
        topic=data.get('topic', 1),
        notes=data.get('notes', 0))

    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json

    else:
        # if this URL is not already in the cache, request it
        start_time = time.time()
        log('Getting from {}, "{}"'.format(prepared_url, data))
        response = requests.get(prepared_url)
        # if this URL is not already in the cache, pause, then request it
        # get the response size and the domain, log result
        size_kb = len(response.content) / 1000.
        domain = re.findall(r'//(?s)(.*?)/', prepared_url)[0]
        log('Downloaded {:,.1f}KB from {}'
            ' in {:,.2f} seconds'.format(size_kb, domain,
                                         time.time() - start_time))

        try:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            save_to_cache(prepared_url, response_json)

        except Exception:
            # There seems to be a double backlash in the response. We try
            # removing it here.
            try:
                response = response.content.decode('UTF-8').replace('//',
                                                                    '')
                response_json = json.loads(response)
            except Exception:
                log(
                    'Server at {} returned status code {} and no JSON '
                    'data.'.format(
                        domain,
                        response.status_code),
                    level=lg.ERROR)
            else:
                save_to_cache(prepared_url, response_json)
                return response_json
            # deal with response satus_code here
            log('Server at {} returned status code {} and no JSON '
                'data.'.format(
                domain, response.status_code), level=lg.ERROR)
        else:
            return response_json


def stat_can_geo_request(data):
    """
    Args:
        data:
    """
    prepared_url = 'https://www12.statcan.gc.ca/rest/census-recensement' \
                   '/CR2016Geo.{type}?lang={lang}&geos={geos}&cpt={cpt}'.format(
        type=data.get('type', 'json'),
        lang=data.get('land', 'E'),
        geos=data.get('geos', 'PR'),
        cpt=data.get('cpt', '00'))

    cached_response_json = get_from_cache(prepared_url)

    if cached_response_json is not None:
        # found this request in the cache, just return it instead of making a
        # new HTTP call
        return cached_response_json

    else:
        # if this URL is not already in the cache, request it
        start_time = time.time()
        log('Getting from {}, "{}"'.format(prepared_url, data))
        response = requests.get(prepared_url)
        # if this URL is not already in the cache, pause, then request it
        # get the response size and the domain, log result
        size_kb = len(response.content) / 1000.
        domain = re.findall(r'//(?s)(.*?)/', prepared_url)[0]
        log('Downloaded {:,.1f}KB from {}'
            ' in {:,.2f} seconds'.format(size_kb, domain,
                                         time.time() - start_time))

        try:
            response_json = response.json()
            if 'remark' in response_json:
                log('Server remark: "{}"'.format(response_json['remark'],
                                                 level=lg.WARNING))
            save_to_cache(prepared_url, response_json)

        except Exception:
            # There seems to be a double backlash in the response. We try
            # removing it here.
            try:
                response = response.content.decode('UTF-8').replace('//',
                                                                    '')
                response_json = json.loads(response)
            except Exception:
                log(
                    'Server at {} returned status code {} and no JSON '
                    'data.'.format(
                        domain,
                        response.status_code),
                    level=lg.ERROR)
            else:
                save_to_cache(prepared_url, response_json)
                return response_json
            # deal with response satus_code here
            log('Server at {} returned status code {} and no JSON '
                'data.'.format(
                domain, response.status_code), level=lg.ERROR)
        else:
            return response_json


def download_bld_window(u_factor, shgc, vis_trans, oauth_key, tolerance=0.05,
                        extension='idf', output_folder=None):
    """Find window constructions corresponding to a combination of a u_factor,
    shgc and visible transmittance and download their idf file to disk. it is
    necessary to have an authentication key (see Info below).

    .. _Building_Component_Library: https://bcl.nrel.gov/user/register

    Args:
        u_factor (float or tuple): The center of glass u-factor. Pass a range of
            values by passing a tuple (from, to). If a tuple is passed,
            *tolerance* is ignored.
        shgc (float or tuple): The Solar Heat Gain Coefficient. Pass a range of
            values by passing a tuple (from, to). If a tuple is passed,
            *tolerance* is ignored.
        vis_trans (float or tuple): The Visible Transmittance. Pass a range of
            values by passing a tuple (from, to). If a tuple is passed,
            *tolerance* is ignored.
        oauth_key (str): the Building_Component_Library_ authentication key.
        tolerance (float): relative tolerance for the input values. Default is
            0.05 (5%).
        extension (str): specify the extension of the file to download.
            (default: 'idf')
        output_folder (str, optional): specify folder to save response data to.
            Defaults to settings.data_folder.

    Returns:
        archetypal.IDF: a list of IDF files containing window objects
            matching the parameters.

    Note:
        An authentication key from NREL is required to download building
        components. Register at Building_Component_Library_
    """
    # check if one or multiple values
    if isinstance(u_factor, tuple):
        u_factor_dict = '[{} TO {}]'.format(u_factor[0], u_factor[1])
    else:
        # apply tolerance
        u_factor_dict = '[{} TO {}]'.format(u_factor * (1 - tolerance),
                                            u_factor * (1 + tolerance))
    if isinstance(shgc, tuple):
        shgc_dict = '[{} TO {}]'.format(shgc[0], shgc[1])
    else:
        # apply tolerance
        shgc_dict = '[{} TO {}]'.format(shgc * (1 - tolerance),
                                        shgc * (1 + tolerance))
    if isinstance(vis_trans, tuple):
        vis_trans_dict = '[{} TO {}]'.format(vis_trans[0], vis_trans[1])
    else:
        # apply tolerance
        vis_trans_dict = '[{} TO {}]'.format(vis_trans * (1 - tolerance),
                                             vis_trans * (1 + tolerance))

    data = {'keyword': 'Window',
            'format': 'json',
            'f[]': ['fs_a_Overall_U-factor:{}'.format(u_factor_dict),
                    'fs_a_VLT:{}'.format(vis_trans_dict),
                    'fs_a_SHGC:{}'.format(shgc_dict),
                    'sm_component_type:"Window"'],
            'oauth_consumer_key': oauth_key}
    response = nrel_bcl_api_request(data)

    if response['result']:
        log('found {} possible window component(s) matching '
            'the range {}'.format(len(response['result']), str(data['f[]'])))

    # download components
    uids = []
    for component in response['result']:
        uids.append(component['component']['uid'])
    url = 'https://bcl.nrel.gov/api/component/download?uids={}'.format(','
                                                                       ''.join(
        uids))
    # actual download with get()
    d_response = requests.get(url)

    if d_response.ok:
        # loop through files and extract the ones that match the extension
        # parameter
        results = []
        if output_folder is None:
            output_folder = settings.data_folder
        with zipfile.ZipFile(io.BytesIO(d_response.content)) as z:
            for info in z.infolist():
                if info.filename.endswith(extension):
                    z.extract(info, path=output_folder)
                    results.append(os.path.join(settings.data_folder,
                                                info.filename))
        return results
    else:
        return response['result']
